# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> "self" data import
"""

import logging
from collections import OrderedDict

from corepos.db.office_op import model as corepos

from rattail import importing
from rattail_corepos.corepos.office.importing import db as corepos_importing
from rattail_corepos.corepos.office.importing.db.corepos import FromCoreHandler
from rattail_corepos.corepos.office.util import get_blueline_template, make_blueline


log = logging.getLogger(__name__)


class FromCoreOfficeToSelf(FromCoreHandler, corepos_importing.model.ToCoreHandler):
    """
    Common base class for import handlers which read data from the
    CORE Office DB for the sake of updating misc. other tables in that
    same DB.
    """
    local_key = 'self'

    def begin_local_transaction(self):
        self.session = self.host_session

    def rollback_transaction(self):
        self.rollback_host_transaction()

    def commit_transaction(self):
        self.commit_host_transaction()

    def get_importers(self):
        importers = OrderedDict()
        importers['CustomerClassic'] = CustomerClassicImporter
        return importers


class FromCoreOffice(importing.FromSQLAlchemy):
    """
    Common base class for the "host" side of importers which read data
    from the CORE Office DB for the sake of updating misc. other
    tables in that same DB.
    """


class CustomerClassicImporter(FromCoreOffice, corepos_importing.model.CustomerClassicImporter):
    """
    custdata -> custdata

    Primarily used to update the ``blueLine`` field.
    """
    host_model_class = corepos.CustomerClassic
    supported_fields = [
        'id',
        'blue_line',
    ]
    allow_create = False
    allow_delete = False

    def setup(self):
        super().setup()
        self.blueline_template = get_blueline_template(self.config)

    def normalize_host_object(self, customer):

        if not customer.member_type:
            log.warning("skipping customer #%s with no member type: %s",
                        customer.card_number, customer)
            return

        blueline = make_blueline(self.config, customer, template=self.blueline_template)
        return {
            'id': customer.id,
            'blue_line': blueline,
        }



# -*- coding: utf-8; -*-
"""add customer.card_number

Revision ID: ae74c537ea51
Revises: d6a0f21a6a94
Create Date: 2023-06-05 19:04:25.574077

"""

# revision identifiers, used by Alembic.
revision = 'ae74c537ea51'
down_revision = 'd6a0f21a6a94'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_customer
    op.alter_column('corepos_customer', 'corepos_account_id',
               existing_type=sa.INTEGER(),
               nullable=True)
    op.add_column('corepos_customer', sa.Column('corepos_card_number', sa.Integer(), nullable=True))
    op.add_column('corepos_customer_version', sa.Column('corepos_card_number', sa.Integer(), autoincrement=False, nullable=True))


def downgrade():

    # corepos_customer
    op.drop_column('corepos_customer_version', 'corepos_card_number')
    op.drop_column('corepos_customer', 'corepos_card_number')
    op.alter_column('corepos_customer', 'corepos_account_id',
               existing_type=sa.INTEGER(),
               nullable=False)



# -*- coding: utf-8; -*-
"""add core_member.corepos_card_number

Revision ID: 15bf65f68c52
Revises: 673ff7088d35
Create Date: 2023-10-12 07:39:34.608105

"""

# revision identifiers, used by Alembic.
revision = '15bf65f68c52'
down_revision = '673ff7088d35'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_member
    op.alter_column('corepos_member', 'corepos_account_id',
               existing_type=sa.INTEGER(),
               nullable=True)
    op.add_column('corepos_member', sa.Column('corepos_card_number', sa.Integer(), nullable=True))
    op.add_column('corepos_member_version', sa.Column('corepos_card_number', sa.Integer(), autoincrement=False, nullable=True))

def downgrade():

    # corepos_member
    op.drop_column('corepos_member_version', 'corepos_card_number')
    op.drop_column('corepos_member', 'corepos_card_number')
    op.alter_column('corepos_member', 'corepos_account_id',
               existing_type=sa.INTEGER(),
               nullable=False)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for CORE equity import batches
"""

import datetime

from corepos.db.office_trans import model as coretrans, Session as CoreTransSession

from rattail.batch import BatchHandler
from rattail_corepos.db.model import CoreEquityImportBatch


class CoreEquityImportBatchHandler(BatchHandler):
    """
    Handler for CORE member batches.
    """
    batch_model_class = CoreEquityImportBatch

    pseudo_remove_rows = False

    def refresh_row(self, row):
        session = self.app.get_session(row)
        model = self.model

        row.status_code = None
        row.status_text = None

        if not row.card_number:
            row.status_code = row.STATUS_MISSING_VALUES
            row.status_text = "card_number"
            return

        if not row.payment_amount:
            row.status_code = row.STATUS_MISSING_VALUES
            row.status_text = "payment_amount"
            return

        if not row.department_number:
            row.status_code = row.STATUS_MISSING_VALUES
            row.status_text = "department_number"
            return

        if not row.timestamp:
            row.status_code = row.STATUS_MISSING_VALUES
            row.status_text = "timestamp"
            return

        payment = row.payment
        if payment and payment.corepos_transaction_number:
            row.status_code = row.STATUS_ALREADY_IN_CORE
            return

        member = payment.member if payment else None
        row.member = member
        if not member:
            row.status_code = row.STATUS_MEMBER_NOT_FOUND
            return

        memtype = member.membership_type
        row.member_type_id = memtype.number if memtype else None

        if member.person:
            person = member.person
            row.first_name = person.first_name
            row.last_name = person.last_name

        membership = self.app.get_membership_handler()
        row.rattail_equity_total = membership.get_equity_total(member)

        row.status_code = row.STATUS_OK

    def describe_execution(self, batch, **kwargs):
        return "New payment transactions will be added directly to CORE-POS."

    def get_effective_rows(self, batch):
        return [row for row in batch.active_rows()
                if row.status_code not in (row.STATUS_MISSING_VALUES,
                                           row.STATUS_MEMBER_NOT_FOUND,
                                           row.STATUS_ALREADY_IN_CORE)]

    def execute(self, batch, progress=None, **kwargs):
        rows = self.get_effective_rows(batch)
        self.export_payments_to_corepos(rows, progress=progress)
        return True

    def get_store_id(self, row):
        # TODO: what should this be?
        return 1

    def get_register_number(self, row):
        # TODO: what should this be?
        return 1

    def get_employee_number(self, row):
        # TODO: what should this be?
        return 0

    def get_next_transaction_number(self, session):
        # TODO: how should we generate this?
        return self.consume_batch_id(session)

    def get_next_timestamp(self, row):
        dt = self.next_timestamp
        self.next_timestamp += datetime.timedelta(seconds=1)
        return self.app.localtime(dt, tzinfo=False).replace(microsecond=0)

    def get_transaction_status(self, row):
        # TODO: what should this be?
        return 'G'

    def export_payments_to_corepos(self, rows, progress=None):
        coretrans_session = CoreTransSession()
        self.next_timestamp = self.app.localtime()

        def export(row, i):
            session = self.app.get_session(row)

            # will insert 2 records in `core_trans.dtransactions` ...

            # they will have this data in common
            host_data = {
                'store_id': self.get_store_id(row),
                'register_number': self.get_register_number(row),
                'transaction_number': self.get_next_transaction_number(session),
                'employee_number': self.get_employee_number(row),
                'card_number': row.card_number,
                'date_time': self.get_next_timestamp(row),
                'total': row.payment_amount,
                'transaction_status': self.get_transaction_status(row),
            }

            # first a record to ring up the equity item
            detail = coretrans.TransactionDetail(**host_data)
            detail.transaction_id = 1
            detail.upc = 'TEST_ITEM' # TODO: what should this say?
            detail.description = 'TEST_EQUITY_ITEM' # TODO: what should this say?
            detail.department_number = row.department_number
            detail.quantity = 1 # TODO: should this ever be anything else?
            detail.unit_price = row.payment_amount
            coretrans_session.add(detail)

            # then a record to accept tender payment
            detail = coretrans.TransactionDetail(**host_data)
            detail.transaction_id = 2
            detail.upc = 'TEST_TENDER' # TODO: what should this say?
            detail.description = 'TEST_EQUITY_TENDER' # TODO: what should this say?
            coretrans_session.add(detail)

            # update payment in Rattail to indicate presence in CORE
            payment = row.payment
            if payment:
                # nb. these must match exactly to be used later as importer key
                payment.corepos_card_number = row.card_number
                payment.corepos_department_number = row.department_number
                payment.corepos_transaction_number = f"{host_data['employee_number']}-{host_data['register_number']}-{host_data['transaction_number']}"
                payment.corepos_transaction_id = 1
                payment.corepos_datetime = self.app.make_utc(self.app.localtime(host_data['date_time']))
                # TODO: stop setting these, *after* importer stops using for key
                # (or perhaps config should determine whether to set?)
                payment.received = payment.corepos_datetime
                payment.transaction_identifier = payment.corepos_transaction_number

        self.progress_loop(export, rows, progress,
                           message="Exporting payments to CORE-POS")

        coretrans_session.commit()
        coretrans_session.close()



# -*- coding: utf-8; -*-
"""add equity import batch

Revision ID: 93978a7adc66
Revises: 08d879bbe118
Create Date: 2023-09-12 23:09:09.148879

"""

# revision identifiers, used by Alembic.
revision = '93978a7adc66'
down_revision = '08d879bbe118'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_corepos_equity_import
    op.create_table('batch_corepos_equity_import',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('complete', sa.Boolean(), nullable=False),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True),
                    sa.Column('extra_data', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'], name='batch_corepos_equity_import_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='batch_corepos_equity_import_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'], name='batch_corepos_equity_import_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('batch_corepos_equity_import_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('modified', sa.DateTime(), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('card_number', sa.Integer(), nullable=True),
                    sa.Column('first_name', sa.String(length=30), nullable=True),
                    sa.Column('last_name', sa.String(length=30), nullable=True),
                    sa.Column('member_type_id', sa.Integer(), nullable=True),
                    sa.Column('payment_amount', sa.Numeric(precision=10, scale=2), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('timestamp', sa.DateTime(), nullable=True),
                    sa.Column('corepos_equity_total', sa.Numeric(precision=10, scale=2), nullable=True),
                    sa.Column('rattail_equity_total', sa.Numeric(precision=10, scale=2), nullable=True),
                    sa.Column('other_equity_total', sa.Numeric(precision=10, scale=2), nullable=True),
                    sa.Column('payment_uuid', sa.String(length=32), nullable=True),
                    sa.Column('member_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], ['batch_corepos_equity_import.uuid'], name='batch_corepos_equity_import_row_fk_batch_uuid'),
                    sa.ForeignKeyConstraint(['payment_uuid'], ['member_equity_payment.uuid'], name='batch_corepos_equity_import_row_fk_payment'),
                    sa.ForeignKeyConstraint(['member_uuid'], ['member.uuid'], name='batch_corepos_equity_import_row_fk_member'),
                    sa.PrimaryKeyConstraint('uuid')
                    )


def downgrade():

    # batch_corepos_equity_import
    op.drop_table('batch_corepos_equity_import_row')
    op.drop_table('batch_corepos_equity_import')



# -*- coding: utf-8; -*-
"""add equity_import tender_code

Revision ID: f8df04546a59
Revises: d426a274f0c2
Create Date: 2023-09-16 15:38:25.569334

"""

# revision identifiers, used by Alembic.
revision = 'f8df04546a59'
down_revision = 'd426a274f0c2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_corepos_equity_import_row
    op.add_column('batch_corepos_equity_import_row', sa.Column('tender_code', sa.String(length=2), nullable=True))


def downgrade():

    # batch_corepos_equity_import_row
    op.drop_column('batch_corepos_equity_import_row', 'tender_code')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS commands
"""

import typer
from typing_extensions import Annotated

from rattail.commands.typer import make_typer, importer_command, typer_get_runas_user
from rattail.commands.importing import ImportCommandHandler



crepes_typer = make_typer(
    name='crepes',
    help="Crepes -- command line interface for CORE-POS"
)


@crepes_typer.command()
@importer_command
def export_core(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the \"target\" "
                         "CORE DB, i.e. where data will be exported.  This key must be "
                         "defined in the [rattail_corepos.db] section of your config file.")] = 'host',
        **kwargs
):
    """
    Export data to another CORE database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_corepos_db_office_op.from_corepos_db_office_op.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)


@crepes_typer.command()
@importer_command
def import_core(
        ctx: typer.Context,
        dbkey: Annotated[
            str,
            typer.Option(help="Config key for database engine to be used as the CORE "
                         "\"host\", i.e. the source of the data to be imported.  This key "
                         "must be defined in the [rattail_corepos.db] section of your "
                         "config file.")] = 'host',
        **kwargs
):
    """
    Import data from another CORE database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_corepos_db_office_op.from_corepos_db_office_op.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {'dbkey': dbkey}
    handler.run(kwargs, progress=progress)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE-POS -> Catapult Inventory Workbook
"""

import re
import datetime
import decimal
import logging
from collections import OrderedDict

from sqlalchemy.exc import ProgrammingError
from sqlalchemy import orm
from sqlalchemy.orm.exc import NoResultFound

from corepos import enum as corepos_enum
from corepos.db.office_op import model as corepos
from corepos.db.util import table_exists

from rattail.gpc import GPC
from rattail.core import get_uuid
from rattail.importing.handlers import ToFileHandler
from rattail_corepos.corepos.importing.db.corepos import FromCoreHandler, FromCore
from rattail_onager.catapult.importing import inventory as catapult_importing


log = logging.getLogger(__name__)


class FromCoreToCatapult(FromCoreHandler, ToFileHandler):
    """
    Handler for CORE -> Catapult (Inventory Workbook)
    """
    host_title = "CORE-POS"
    local_title = "Catapult (Inventory Workbook)"
    direction = 'export'

    def get_importers(self):
        importers = OrderedDict()
        importers['InventoryItem'] = InventoryItemImporter
        return importers


class InventoryItemImporter(FromCore, catapult_importing.model.InventoryItemImporter):
    """
    Inventory Item data importer.
    """
    host_model_class = corepos.Product
    # note that we use a "dummy" uuid key here, so logic will consider each row
    # to be unique, even when duplicate item_id's are present
    key = 'uuid'
    supported_fields = [
        'uuid',
        'item_id',
        'dept_id',
        'dept_name',
        'receipt_alias',
        'brand',
        'item_name',
        'size',
        # 'sugg_retail',
        'last_cost',
        'price_divider',
        'base_price',
        # 'disc_mult',
        'ideal_margin',
        'bottle_deposit',
        # 'pos_menu_group',
        'scale_label',
        'sold_by_ea_or_lb',
        'quantity_required',
        'weight_profile',
        'tax_1',
        'tax_2',
        'spec_tend_1',
        'spec_tend_2',
        'age_required',
        'location',
        # 'family_line',
        'alt_id',
        'alt_receipt_alias',
        'alt_pkg_qty',
        'alt_pkg_price',
        'auto_discount',
        'supplier_unit_id',
        'supplier_id',
        'unit',
        'num_pkgs',
        # 'cs_pk_multiplier',
        # 'dsd',
        'pf1',
        # 'pf2',
        # 'pf3',
        # 'pf4',
        # 'pf5',
        # 'pf6',
        # 'pf7',
        # 'pf8',
        'memo',
        'scale_shelf_life',
        'scale_shelf_life_type',
        'scale_ingredient_text',
    ]

    # we want to add a "duplicate" column at the end
    include_duplicate_column = True

    # we want to add an "alternate for" column at the end
    include_alt_for_column = True

    type2_upc_pattern = re.compile(r'^2(\d{5})00000\d')

    def setup(self):
        super(InventoryItemImporter, self).setup()

        # this is used for sorting, when a value has no date
        self.old_datetime = datetime.datetime(1900, 1, 1)

        self.exclude_invalid_upc = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.exclude_invalid_upc',
            default=False)

        self.warn_invalid_upc = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_invalid_upc',
            default=True)

        self.ignored_upcs = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.ignored_upcs')

        self.exclude_missing_department = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.exclude_missing_department',
            default=False)

        self.warn_missing_department = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_missing_department',
            default=True)

        self.warn_empty_subdepartment = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_empty_subdepartment',
            default=True)

        self.warn_truncated_receipt_alias = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_truncated_receipt_alias',
            default=True)

        self.warn_size_null_byte = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_size_null_byte',
            default=True)

        self.warn_unknown_deposit = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_unknown_deposit',
            default=True)

        self.warn_scale_label_non_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_label_non_plu',
            default=True)

        self.warn_scale_label_short_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_label_short_plu',
            default=True)

        self.warn_weight_profile_non_plu = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_weight_profile_non_plu',
            default=True)

        self.warn_multiple_vendor_items = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_multiple_vendor_items',
            default=True)

        self.warn_no_valid_vendor_items = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_no_valid_vendor_items',
            default=True)

        self.warn_truncated_memo = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_truncated_memo',
            default=True)

        self.warn_scale_ingredients_newline = self.config.getbool(
            'corepos', 'exporting.catapult_inventory.warn_scale_ingredients_newline',
            default=True)

        self.floor_sections_exist = table_exists(self.host_session,
                                                 corepos.FloorSection)
        self.tax_components_exist = table_exists(self.host_session,
                                                 corepos.TaxRateComponent)

        self.tax_rate_ids_1 = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.tax_rate_ids_1', default=[])
        self.tax_rate_ids_1 = [int(id) for id in self.tax_rate_ids_1]
        self.tax_rate_ids_2 = self.config.getlist(
            'corepos', 'exporting.catapult_inventory.tax_rate_ids_2', default=[])
        self.tax_rate_ids_2 = [int(id) for id in self.tax_rate_ids_2]

        # TODO: should add component id levels too?
        # tax_component_ids_1 = (1,)
        # tax_component_ids_2 = (2,)

        self.cache_bottle_deposits()
        self.cache_like_codes()

    def cache_bottle_deposits(self):
        self.deposits = {}
        deposits = self.host_session.query(corepos.Product.deposit.distinct())\
                                    .all()

        def cache(deposit, i):
            assert isinstance(deposit, tuple)
            assert len(deposit) == 1
            deposit = deposit[0]
            if deposit:
                deposit = int(deposit)
                upc = "{:013d}".format(deposit)
                try:
                    product = self.host_session.query(corepos.Product)\
                                               .filter(corepos.Product.upc == upc)\
                                               .one()
                except NoResultFound:
                    pass        # we will log warnings per-item later
                else:
                    self.deposits[deposit] = product

        self.progress_loop(cache, deposits,
                           message="Caching product deposits data")

    def cache_like_codes(self):
        self.like_codes = {}
        mappings = self.host_session.query(corepos.ProductLikeCode)\
                                    .order_by(corepos.ProductLikeCode.like_code_id,
                                              corepos.ProductLikeCode.upc)\
                                    .all()

        def cache(mapping, i):
            self.like_codes.setdefault(mapping.like_code_id, []).append(mapping)

        self.progress_loop(cache, mappings,
                           message="Caching like codes data")

    def query(self):
        query = self.host_session.query(corepos.Product)\
                                 .order_by(corepos.Product.upc)\
                                 .options(orm.joinedload(corepos.Product.department))\
                                 .options(orm.joinedload(corepos.Product.subdepartment))\
                                 .options(orm.joinedload(corepos.Product.vendor_items)\
                                          .joinedload(corepos.VendorItem.vendor))\
                                 .options(orm.joinedload(corepos.Product.default_vendor))\
                                 .options(orm.joinedload(corepos.Product.scale_item))\
                                 .options(orm.joinedload(corepos.Product.user_info))\
                                 .options(orm.joinedload(corepos.Product.tax_rate))\
                                 .options(orm.joinedload(corepos.Product._like_code))
        if self.floor_sections_exist:
            query = query.options(orm.joinedload(corepos.Product.physical_location)\
                                  .joinedload(corepos.ProductPhysicalLocation.floor_section))
        return query

    def normalize_host_data(self, host_objects=None):
        normalized = super(InventoryItemImporter, self).normalize_host_data(host_objects=host_objects)

        # re-sort the results by item_id, since e.g. original UPC from CORE may
        # have been replaced with a PLU.  also put non-numeric first, to bring
        # them to user's attention
        numeric = []
        non_numeric = []
        for row in normalized:
            if row['item_id'] and row['item_id'].isdigit():
                numeric.append(row)
            else:
                non_numeric.append(row)
        numeric.sort(key=lambda row: int(row['item_id']))
        non_numeric.sort(key=lambda row: row['item_id'])
        normalized = non_numeric + numeric

        # now we must check for duplicate item ids, and mark rows accordingly.
        # but we *do* want to include/preserve all rows, hence we mark them
        # instead of pruning some out. first step is to group all by item_id
        items = {}

        def collect(row, i):
            items.setdefault(row['item_id'], []).append(row)

        self.progress_loop(collect, normalized,
                           message="Grouping rows by Item ID")

        # now we go through our groupings and for any item_id with more than 1
        # row, we'll mark each row as having a duplicate item_id.  note that
        # this modifies such a row "in-place" for our overall return value
        def inspect(rows, i):
            if len(rows) > 1:
                for row in rows:
                    row['__duplicate__'] = True

        self.progress_loop(inspect, list(items.values()),
                           message="Marking any duplicate Item IDs")

        # finally, we must inspect the like codes and figure out which
        # product(s) should potentially be considered "alternate for" another.
        # first step here will be to create mapping of item_id values for each
        # CORE product in our result set
        item_ids = {}

        def mapp(row, i):
            product = row['__product__']
            item_ids[product.upc] = row['item_id']

        self.progress_loop(mapp, normalized,
                           message="Mapping item_id for CORE products")

        # next step here is to check each product and mark "alt for" as needed
        def inspect(row, i):
            product = row['__product__']
            if product.like_code:
                others = self.like_codes.get(product.like_code.id)
                if others:
                    first = others[0]
                    if first.upc != product.upc:
                        row['__alternate_for__'] = item_ids[first.upc]

        self.progress_loop(inspect, normalized,
                           message="Marking any \"alternate for\" items")

        return normalized

    def normalize_host_object(self, product):
        item_id = product.upc

        if self.ignored_upcs and item_id in self.ignored_upcs:
            log.debug("ignoring UPC %s for product: %s", product.upc, product)
            return

        if not item_id:
            logger = log.warning if self.warn_invalid_upc else log.debug
            logger("product id %s has no upc: %s", product.id, product)
            if self.exclude_invalid_upc:
                return

        if not item_id.isdigit():
            logger = log.warning if self.warn_invalid_upc else log.debug
            logger("product %s has non-numeric upc: %s",
                   product.upc, product)
            if self.exclude_invalid_upc:
                return

        # convert item_id either to a PLU, or formatted UPC
        is_plu = False
        if item_id.isdigit():   # can only convert if it's numeric!
            if len(str(int(item_id))) < 6:
                is_plu = True
                item_id = str(int(item_id))
            else: # must add check digit, and re-format
                upc = GPC(item_id, calc_check_digit='upc')
                item_id = str(upc)
                assert len(item_id) == 14
                # drop leading zero(s)
                if item_id[1] == '0': # UPC-A
                    item_id = item_id[2:]
                    assert len(item_id) == 12
                else: # EAN13
                    item_id = item_id[1:]
                    assert len(item_id) == 13

        # figure out the "scale label" data, which may also affect item_id
        scale_item = product.scale_item
        scale_label = None
        if scale_item:
            scale_label = 'Y'
            if item_id.isdigit():
                if len(item_id) < 5:
                    logger = log.warning if self.warn_scale_label_short_plu else log.debug
                    logger("product %s has scale label, but PLU is less than 5 digits (%s): %s",
                           product.upc, item_id, product)
                elif len(item_id) > 5:
                    match = self.type2_upc_pattern.match(item_id)
                    if match:
                        # convert type-2 UPC to PLU
                        is_plu = True
                        item_id = str(int(match.group(1)))
                        log.debug("converted type-2 UPC %s to PLU %s for: %s",
                                  product.upc, item_id, product)
                    else:
                        logger = log.warning if self.warn_scale_label_non_plu else log.debug
                        logger("product %s has scale label, but non-PLU item_id: %s",
                               product.upc, product)

        department = product.department
        if not department:
            logger = log.warning if self.warn_missing_department else log.debug
            logger("product %s has no department: %s", product.upc, product)
            if self.exclude_missing_department:
                return

        # size may come from one of two fields, or combination thereof
        pack_size = (product.size or '').strip()
        uom = (product.unit_of_measure or '').strip()
        numeric_pack = False
        if pack_size:
            try:
                decimal.Decimal(pack_size)
            except decimal.InvalidOperation:
                pass
            else:
                numeric_pack = True
        if numeric_pack:
            size = "{} {}".format(pack_size, uom).strip()
        else:
            size = pack_size or uom or None
        # TODO: this logic may actually be client-specific?  i just happened to
        # find some null chars in a client DB and needed to avoid them, b/c the
        # openpyxl lib said IllegalCharacterError
        if size is not None and '\x00' in size:
            logger = log.warning if self.warn_size_null_byte else log.debug
            logger("product %s has null byte in size field: %s",
                   product.upc, product)
            size = size.replace('\x00', '')

        price_divider = None
        if (product.quantity and product.group_price and
            product.price_method == corepos_enum.PRODUCT_PRICE_METHOD_ALWAYS):
            diff = (product.quantity * product.normal_price) - product.group_price
            if abs(round(diff, 2)) > .01:
                log.warning("product %s has multi-price with $%0.2f diff: %s",
                            product.upc, diff, product)
            price_divider = product.quantity

        bottle_deposit = None
        if product.deposit:
            deposit = int(product.deposit)
            if deposit in self.deposits:
                bottle_deposit = self.deposits[deposit].normal_price
            else:
                logger = log.warning if self.warn_unknown_deposit else log.debug
                logger("product %s has unknown deposit %s which will be ignored: %s",
                       product.upc, deposit, product)

        sold_by_ea_or_lb = None
        if is_plu:
            sold_by_ea_or_lb = 'LB' if product.scale else 'EA'

        weight_profile = None
        if product.scale or scale_item:
            if not is_plu:
                logger = log.warning if self.warn_weight_profile_non_plu else log.debug
                logger("product %s has weight profile, but non-PLU item_id %s: %s",
                       product.upc, item_id, product)
            weight_profile = 'LBNT'

        # calculate tax rates according to configured "mappings"
        tax_1 = 0
        tax_2 = 0
        if product.tax_rate:

            # TODO: need to finish the logic to handle tax components
            if self.tax_components_exist and product.tax_rate.components:
                # for component in product.tax_rate.components:
                #     if tax_component_ids_1 and component.id in tax_component_ids_1:
                #         tax_1 += component.rate
                #     if tax_component_ids_2 and component.id in tax_component_ids_2:
                #         tax_2 += component.rate
                raise NotImplementedError

            else: # no components
                rate = product.tax_rate
                if self.tax_rate_ids_1 and rate.id in self.tax_rate_ids_1:
                    tax_1 += rate.rate
                if self.tax_rate_ids_2 and rate.id in self.tax_rate_ids_2:
                    tax_2 += rate.rate
                if not (self.tax_rate_ids_1 or self.tax_rate_ids_2) and rate.rate:
                    log.warning("product %s has unknown tax rate %s (%s) which will "
                                "be considered as tax 1: %s",
                                product.upc, rate.rate, rate.description, product)
                    tax_1 += rate.rate

        location = None
        if self.floor_sections_exist and product.physical_location and product.physical_location.floor_section:
            location = product.physical_location.floor_section.name
            if len(location) > 30:
                log.warning("product %s has location length %s; will truncate: %s",
                            product.upc, len(location), location)
                location = location[:30]

        # no alt item (or auto discount) by default
        alt_id = None
        alt_receipt_alias = None
        alt_pkg_qty = None
        alt_pkg_price = None
        auto_discount = None

        # make an alt item, when main item has pack pricing (e.g. Zevia sodas)
        # note that in this case the main item_id and alt_id are the same
        if (product.quantity and product.group_price and
            product.price_method == corepos_enum.PRODUCT_PRICE_METHOD_FULL_SETS):
            alt_id = item_id
            suffix = "{}-PK".format(product.quantity)
            alt_receipt_alias = "{} {}".format(product.description, suffix)
            if len(alt_receipt_alias) > 32:
                logger = log.warning if self.warn_truncated_receipt_alias else log.debug
                logger("alt receipt alias for %s is %s chars; must truncate: %s",
                       alt_id, len(alt_receipt_alias), alt_receipt_alias)
                overage = len(alt_receipt_alias) - 32
                alt_receipt_alias = "{} {}".format(
                    product.description[:-overage], suffix)
                assert len(alt_receipt_alias) == 32
            alt_pkg_qty = product.quantity
            alt_pkg_price = product.group_price

            # we also must declare an "auto discount" to get pack price
            auto_discount = "{} @ ${:0.2f}".format(alt_pkg_qty, alt_pkg_price)

        # no supplier info by default
        supplier_unit_id = None
        supplier_id = None
        supplier_unit = None
        supplier_num_pkgs = None

        # maybe add supplier info, for "default" `vendorItems` record.  we'll
        # have to get a little creative to figure out which is the default
        vendor_items = []

        # first we try to narrow down according to product's default vendor
        if product.default_vendor:
            vendor_items = [item for item in product.vendor_items
                            if item.vendor is product.default_vendor]

        # but if that didn't work, just use any "valid" vendorItems
        if not vendor_items:
            # valid in this context means, not missing vendor
            vendor_items = [item for item in product.vendor_items
                            if item.vendor]
            if not vendor_items and product.vendor_items:
                logger = log.warning if self.warn_no_valid_vendor_items else log.debug
                logger("product %s has %s vendorItems but each is missing (valid) vendor: %s",
                       product.upc, len(product.vendor_items), product)

        if vendor_items:
            if len(vendor_items) > 1:

                # try to narrow down a bit further, based on valid 'units' amount
                valid_items = [item for item in vendor_items
                               if item.units]
                if valid_items:
                    vendor_items = valid_items

                # warn if we still have more than one "obvious" vendor item
                if len(vendor_items) > 1:
                    logger = log.warning if self.warn_multiple_vendor_items else log.debug
                    logger("product %s has %s vendorItems to pick from: %s",
                           product.upc, len(vendor_items), product)

                    # sort the list so most recently modified is first
                    vendor_items.sort(key=lambda item: item.modified or self.old_datetime,
                                      reverse=True)

            # use the "first" vendor item available
            item = vendor_items[0]
            supplier_unit_id = item.sku
            supplier_id = item.vendor.name
            supplier_num_pkgs = item.units or 1
            if supplier_num_pkgs == 1:
                supplier_unit = 'LB' if product.scale else 'EA'
            else:
                supplier_unit = 'CS'

        pf1 = None
        if product.subdepartment:
            if not product.subdepartment.number:
                logger = log.warning if self.warn_empty_subdepartment else log.debug
                logger("product %s has 'empty' subdepartment number: %s",
                       product.upc, product)
            else:
                pf1 = "{} {}".format(product.subdepartment.number,
                                     product.subdepartment.name)

        memo = None
        if product.user_info and product.user_info.long_text is not None:
            memo = str(product.user_info.long_text)
        if memo and len(memo) > 254:
            logger = log.warning if self.warn_truncated_memo else log.debug
            logger("product %s has memo of length %s; will truncate: %s",
                   product.upc, len(memo), memo)
            memo = memo[:254]

        scale_ingredient_text = None
        if scale_item:
            scale_ingredient_text = scale_item.text
            if "\n" in scale_ingredient_text:
                logger = log.warning if self.warn_scale_ingredients_newline else log.debug
                logger("must remove carriage returns for scale ingredients: %s",
                       scale_ingredient_text)
                scale_ingredient_text = scale_ingredient_text.replace("\n", " ")

        return {
            '__product__': product,
            '__original_item_id__': product.upc,
            'uuid': get_uuid(),
            'item_id': item_id,
            'dept_id': department.number if department else None,
            'dept_name': department.name if department else None,
            'receipt_alias': product.description,
            'brand': product.brand,
            'item_name': product.description,
            'size': size,

            # TODO: does CORE have this?
            # 'sugg_retail': None,

            'last_cost': product.cost,
            'price_divider': price_divider,
            'base_price': product.normal_price,
            'ideal_margin': department.margin * 100 if department and department.margin else None,

            # TODO: does CORE have these?
            # 'disc_mult': None,

            'bottle_deposit': bottle_deposit,

            # TODO: does CORE have this?
            # 'pos_menu_group': None,

            'scale_label': scale_label,
            'sold_by_ea_or_lb': sold_by_ea_or_lb,
            'quantity_required': 'Y' if product.quantity_enforced else None,
            'weight_profile': weight_profile,
            'tax_1': tax_1 or None, # TODO: logic above is unfinished
            'tax_2': tax_2 or None, # TODO: logic above is unfinished
            'spec_tend_1': 'EBT' if product.foodstamp else None,
            'spec_tend_2': 'WIC' if product.wicable else None,
            'age_required': product.id_enforced or None,
            'location': location,

            # TODO: does CORE have these?
            # 'family_line': None,

            'alt_id': alt_id,
            'alt_receipt_alias': alt_receipt_alias,
            'alt_pkg_qty': alt_pkg_qty,
            'alt_pkg_price': alt_pkg_price,
            'auto_discount': auto_discount,
            'supplier_unit_id': supplier_unit_id,
            'supplier_id': supplier_id,
            'unit': supplier_unit,
            'num_pkgs': supplier_num_pkgs,

            # TODO: does CORE have these?
            # 'cs_pk_multiplier': None,
            # 'dsd': None,

            'pf1': pf1,

            # TODO: are these needed?
            # 'pf2',
            # 'pf3',
            # 'pf4',
            # 'pf5',
            # 'pf6',
            # 'pf7',
            # 'pf8',

            'memo': memo,
            'scale_shelf_life': scale_item.shelf_life if scale_item else None,
            'scale_shelf_life_type': 0 if scale_item else None,
            'scale_ingredient_text': scale_ingredient_text,
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database schema extensions for CORE-POS integration
"""

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model


class CorePerson(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Person`.
    """
    __tablename__ = 'corepos_person'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['person.uuid'],
                                name='corepos_person_fk_person'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    person = orm.relationship(
        model.Person,
        doc="""
        Reference to the actual person record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this person.
            """))

    corepos_customer_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``Customers.customerID`` value for this person, within CORE-POS.
    """)

    def __str__(self):
        return str(self.person)

CorePerson.make_proxy(model.Person, '_corepos', 'corepos_customer_id')


class CoreEmployee(model.Base):
    """
    CORE-specific extensions to :class:`~rattail:rattail.db.model.Employee`
    """
    __tablename__ = 'corepos_employee'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['employee.uuid'],
                                name='corepos_employee_fk_employee'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    employee = orm.relationship(
        model.Employee,
        doc="""
        Reference to the actual employee record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this employee.
            """))

    corepos_number = sa.Column(sa.Integer(), nullable=True, doc="""
    ``employees.emp_no`` value for this employee, within CORE-POS.
    """)

    def __str__(self):
        return str(self.employee)

CoreEmployee.make_proxy(model.Employee, '_corepos', 'corepos_number')


class CoreCustomer(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Customer`.
    """
    __tablename__ = 'corepos_customer'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['customer.uuid'],
                                name='corepos_customer_fk_customer'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    customer = orm.relationship(
        model.Customer,
        doc="""
        Reference to the actual customer record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this customer.
            """))

    corepos_account_id = sa.Column(sa.Integer(), nullable=True, doc="""
    ``CustomerAccounts.customerAccountID`` value for this customer
    account, within CORE-POS.
    """)

    corepos_card_number = sa.Column(sa.Integer(), nullable=True, doc="""
    ``custdata.CardNo`` value for this customer, within CORE-POS.
    """)

    def __str__(self):
        return str(self.customer)

CoreCustomer.make_proxy(model.Customer, '_corepos', 'corepos_account_id')
CoreCustomer.make_proxy(model.Customer, '_corepos', 'corepos_card_number')


class CoreCustomerShopper(model.Base):
    """
    CORE-specific extensions to
    :class:`~rattail:rattail.db.model.CustomerShopper`.
    """
    __tablename__ = 'corepos_customer_shopper'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['customer_shopper.uuid'],
                                name='corepos_customer_shopper_fk_shopper'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    shopper = orm.relationship(
        model.CustomerShopper, doc="""
        Reference to the actual shopper record, which this one extends.
        """,
        cascade_backrefs=False,
        backref=orm.backref(
            '_corepos', doc="""
            Reference to the CORE-POS extension record for this customer.
            """,
            uselist=False,
            cascade='all, delete-orphan',
            cascade_backrefs=False))

    # please note, there is *not* a unique constraint on this field.
    # that is intentional, for now, to give some breathing room for
    # testing etc.
    corepos_customer_id = sa.Column(sa.Integer(), nullable=True, doc="""
    ``Customers.customerID`` value for this shopper, within CORE-POS.
    """)

    def __str__(self):
        return str(self.shopper)

CoreCustomerShopper.make_proxy(model.CustomerShopper, '_corepos', 'corepos_customer_id')


class CoreMember(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Member`.
    """
    __tablename__ = 'corepos_member'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['member.uuid'],
                                name='corepos_member_fk_member'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    member = orm.relationship(
        model.Member,
        doc="""
        Reference to the actual member record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this member.
            """))

    corepos_account_id = sa.Column(sa.Integer(), nullable=True, doc="""
    ``Customers.customerAccountID`` value for this member, within CORE-POS.
    """)

    corepos_card_number = sa.Column(sa.Integer(), nullable=True, doc="""
    ``meminfo.card_no`` / ``CustomerAccounts.cardNo`` value for this
    member, within CORE-POS.
    """)

    def __str__(self):
        return str(self.member)

CoreMember.make_proxy(model.Member, '_corepos', 'corepos_account_id')
CoreMember.make_proxy(model.Member, '_corepos', 'corepos_card_number')


class CoreMemberEquityPayment(model.Base):
    """
    CORE-specific extensions to
    :class:`~rattail:rattail.db.model.MemberEquityPayment`.
    """
    __tablename__ = 'corepos_member_equity_payment'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['member_equity_payment.uuid'],
                                name='corepos_member_equity_payment_fk_payment'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    payment = orm.relationship(
        model.MemberEquityPayment,
        doc="""
        Reference to the actual payment record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this payment.
            """))

    corepos_card_number = sa.Column(sa.Integer(), nullable=False, doc="""
    ``stockpurchases.card_no`` value for this payment, within CORE-POS.
    """)

    corepos_transaction_number = sa.Column(sa.String(length=50), nullable=True, doc="""
    ``stockpurchases.trans_num`` value for this payment, within CORE-POS.
    """)

    corepos_transaction_id = sa.Column(sa.Integer(), nullable=True, doc="""
    ``stockpurchases.trans_id`` value for this payment, within CORE-POS.
    """)

    corepos_department_number = sa.Column(sa.Integer(), nullable=True, doc="""
    ``stockpurchases.dept`` value for this payment, within CORE-POS.
    """)

    corepos_datetime = sa.Column(sa.DateTime(), nullable=True, doc="""
    ``stockpurchases.datetime`` value for this payment, within CORE-POS.
    """)

    def __str__(self):
        return str(self.payment)

CoreMemberEquityPayment.make_proxy(model.MemberEquityPayment, '_corepos', 'corepos_card_number')
CoreMemberEquityPayment.make_proxy(model.MemberEquityPayment, '_corepos', 'corepos_transaction_number')
CoreMemberEquityPayment.make_proxy(model.MemberEquityPayment, '_corepos', 'corepos_transaction_id')
CoreMemberEquityPayment.make_proxy(model.MemberEquityPayment, '_corepos', 'corepos_department_number')
CoreMemberEquityPayment.make_proxy(model.MemberEquityPayment, '_corepos', 'corepos_datetime')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail Commands
"""

import typer
from typing_extensions import Annotated

from corepos.enum import CoreDbType

from rattail.commands import rattail_typer
from rattail.commands.typer import importer_command, file_importer_command, typer_get_runas_user
from rattail.commands.importing import ImportCommandHandler


@rattail_typer.command()
@file_importer_command
def corepos_import_square(
        ctx: typer.Context,
        **kwargs
):
    """
    Import transaction data from Square into CORE
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_corepos_db_office_trans.from_square_csv.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def export_corepos(
        ctx: typer.Context,
        **kwargs
):
    """
    Export data from Rattail to CORE-POS
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_corepos_api.from_rattail.export')
    kwargs['user'] = typer_get_runas_user(ctx)
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def import_corepos_api(
        ctx: typer.Context,
        **kwargs
):
    """
    Import data from a CORE POS API
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_corepos_api.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    handler.run(kwargs, progress=progress)


@rattail_typer.command()
@importer_command
def import_corepos_db(
        ctx: typer.Context,
        corepos_dbtype: Annotated[
            CoreDbType,
            typer.Option(help="Type of CORE-POS DB engine to be used as data host.  "
                         "This determines which config section is used with regard "
                         "to the --corepos-dbkey arg.")] = 'office_op',
        corepos_dbkey: Annotated[
            str,
            typer.Option(help="Config key for CORE POS database engine to be used as "
                         "the \"host\", i.e. the source of the data to be imported.  "
                         "This key must be defined in the [rattail_corepos.db] section "
                         "of your config file.")] = 'default',
        **kwargs
):
    """
    Import data from a CORE POS database
    """
    config = ctx.parent.rattail_config
    progress = ctx.parent.rattail_progress
    handler = ImportCommandHandler(
        config, import_handler_key='to_rattail.from_corepos_db_office_op.import')
    kwargs['user'] = typer_get_runas_user(ctx)
    kwargs['handler_kwargs'] = {
        'corepos_dbtype': corepos_dbtype,
        'corepos_dbkey': corepos_dbkey,
    }
    handler.run(kwargs, progress=progress)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database schema extensions for CORE-POS integration
"""

from .stores import CoreStore, CoreTax, CoreTender
from .people import (CorePerson, CoreEmployee,
                     CoreCustomer, CoreCustomerShopper,
                     CoreMember, CoreMemberEquityPayment)
from .products import (CoreDepartment, CoreSubdepartment,
                       CoreVendor, CoreProduct, CoreProductCost)

from .batch.coremember import CoreMemberBatch, CoreMemberBatchRow
from .batch.equityimport import CoreEquityImportBatch, CoreEquityImportBatchRow



# -*- coding: utf-8; -*-
"""add core_employee

Revision ID: 1f2e2f57c90b
Revises: c40a6ec00428
Create Date: 2023-10-01 19:03:56.921897

"""

# revision identifiers, used by Alembic.
revision = '1f2e2f57c90b'
down_revision = 'c40a6ec00428'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_employee
    op.create_table('corepos_employee',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_number', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['uuid'], ['employee.uuid'], name='corepos_employee_fk_employee'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('corepos_employee_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_corepos_employee_version_end_transaction_id'), 'corepos_employee_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_employee_version_operation_type'), 'corepos_employee_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_employee_version_transaction_id'), 'corepos_employee_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_employee
    op.drop_index(op.f('ix_corepos_employee_version_transaction_id'), table_name='corepos_employee_version')
    op.drop_index(op.f('ix_corepos_employee_version_operation_type'), table_name='corepos_employee_version')
    op.drop_index(op.f('ix_corepos_employee_version_end_transaction_id'), table_name='corepos_employee_version')
    op.drop_table('corepos_employee_version')
    op.drop_table('corepos_employee')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office - data anonymization
"""

import random

from rattail.app import GenericHandler
from rattail.db.util import finalize_session
from rattail_corepos.corepos.office.util import get_blueline_template, make_blueline


class Anonymizer(GenericHandler):
    """
    Make anonymous (randomize) all customer names etc.
    """

    def anonymize_all(self, dbkey=None, dry_run=False, progress=None):
        import names
        import us

        core_handler = self.app.get_corepos_handler()
        op_session = core_handler.make_session_office_op(dbkey=dbkey)
        op_model = core_handler.get_model_office_op()

        states = [state.abbr for state in us.states.STATES]

        # meminfo
        members = op_session.query(op_model.MemberInfo).all()
        members_by_card_number = {}

        def anon_meminfo(member, i):
            member.first_name = names.get_first_name()
            member.last_name = names.get_last_name()
            member.other_first_name = names.get_first_name()
            member.other_last_name = names.get_last_name()
            member.street = '123 Main St.'
            member.city = 'Anytown'
            member.state = random.choice(states)
            member.zipcode = self.random_zipcode()
            member.phone = self.random_phone()
            member.email = self.random_email()
            member.notes.clear()
            members_by_card_number[member.card_number] = member

        self.app.progress_loop(anon_meminfo, members, progress,
                               message="Anonymizing meminfo")

        # custdata
        customers = op_session.query(op_model.CustomerClassic).all()
        blueline_template = get_blueline_template(self.config)

        def anon_custdata(customer, i):
            member = members_by_card_number.get(customer.card_number)
            if member:
                customer.first_name = member.first_name
                customer.last_name = member.last_name
            else:
                customer.first_name = names.get_first_name()
                customer.last_name = names.get_last_name()
            customer.blue_line = make_blueline(self.config, customer,
                                               template=blueline_template)

        self.app.progress_loop(anon_custdata, customers, progress,
                               message="Anonymizing custdata")

        # Customers
        customers = op_session.query(op_model.Customer).all()

        def del_customer(customer, i):
            op_session.delete(customer)

        self.app.progress_loop(del_customer, customers, progress,
                               message="Deleting from Customers")

        # CustomerAccounts
        accounts = op_session.query(op_model.CustomerAccount).all()

        def del_account(account, i):
            op_session.delete(account)

        self.app.progress_loop(del_account, accounts, progress,
                               message="Deleting from CustomerAccounts")

        # employees
        employees = op_session.query(op_model.Employee).all()

        def anon_employee(employee, i):
            employee.first_name = names.get_first_name()
            employee.last_name = names.get_last_name()

        self.app.progress_loop(anon_employee, employees, progress,
                               message="Anonymizing employees")

        # Users
        users = op_session.query(op_model.User).all()

        def anon_user(user, i):
            user.real_name = names.get_full_name()

        self.app.progress_loop(anon_user, users, progress,
                               message="Anonymizing users")

        finalize_session(op_session, dry_run=dry_run)

    def random_phone(self):
        digits = [random.choice('0123456789')
                  for i in range(10)]
        return self.app.format_phone_number(''.join(digits))

    def random_email(self):
        import names
        name = names.get_full_name()
        name = name.replace(' ', '_')
        return f'{name}@mailinator.com'

    def random_zipcode(self):
        digits = [random.choice('0123456789')
                  for i in range(5)]
        return ''.join(digits)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common importing logic for CORE-POS
"""

from rattail import importing


class ToCore(importing.ToSQLAlchemy):
    """
    Base class for all CORE model importers; i.e. anything which uses
    a CORE DB for the local side.
    """

    def create_object(self, key, host_data):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return host_data

        return super(ToCore, self).create_object(key, host_data)

    def update_object(self, obj, host_data, **kwargs):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return obj

        return super(ToCore, self).update_object(obj, host_data, **kwargs)

    def delete_object(self, obj):

        # NOTE! some tables in CORE DB may be using the MyISAM storage engine,
        # which means it is *not* transaction-safe and therefore we cannot rely
        # on "rollback" if in dry-run mode!  in other words we better not touch
        # the record at all, for dry run
        if self.dry_run:
            return True

        return super(ToCore, self).delete_object(obj)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database schema extensions for CORE-POS integration
"""

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model


class CoreDepartment(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Department`.
    """
    __tablename__ = 'corepos_department'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['department.uuid'],
                                name='corepos_department_fk_department'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    department = orm.relationship(
        model.Department,
        doc="""
        Reference to the actual department record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this department.
            """))

    corepos_number = sa.Column(sa.Integer(), nullable=False, doc="""
    ``dept_no`` value for the department, within CORE-POS.
    """)

    def __str__(self):
        return str(self.department)

CoreDepartment.make_proxy(model.Department, '_corepos', 'corepos_number')


class CoreSubdepartment(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Subdepartment`.
    """
    __tablename__ = 'corepos_subdepartment'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['subdepartment.uuid'],
                                name='corepos_subdepartment_fk_subdepartment'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    subdepartment = orm.relationship(
        model.Subdepartment,
        doc="""
        Reference to the actual subdepartment record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this subdepartment.
            """))

    corepos_number = sa.Column(sa.Integer(), nullable=False, doc="""
    ``subdept_no`` value for the subdepartment, within CORE-POS.
    """)

    def __str__(self):
        return str(self.subdepartment)

CoreSubdepartment.make_proxy(model.Subdepartment, '_corepos', 'corepos_number')


class CoreVendor(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Vendor`.
    """
    __tablename__ = 'corepos_vendor'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['vendor.uuid'],
                                name='corepos_vendor_fk_vendor'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    vendor = orm.relationship(
        model.Vendor,
        doc="""
        Reference to the actual vendor record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this vendor.
            """))

    corepos_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``vendorID`` value for the vendor, within CORE-POS.
    """)

    def __str__(self):
        return str(self.vendor)

CoreVendor.make_proxy(model.Vendor, '_corepos', 'corepos_id')


class CoreProduct(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Product`.
    """
    __tablename__ = 'corepos_product'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['product.uuid'],
                                name='corepos_product_fk_product'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    product = orm.relationship(
        model.Product,
        doc="""
        Reference to the actual product record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this product.
            """))

    corepos_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``id`` value for the product, within CORE-POS.
    """)

    def __str__(self):
        return str(self.product)

CoreProduct.make_proxy(model.Product, '_corepos', 'corepos_id')


class CoreProductCost(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.ProductCost`.
    """
    __tablename__ = 'corepos_product_cost'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['product_cost.uuid'],
                                name='corepos_product_cost_fk_cost'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    cost = orm.relationship(
        model.ProductCost,
        doc="""
        Reference to the actual ProductCost record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this product cost.
            """))

    corepos_vendor_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``vendorItems.vendorID`` value for the corresponding record within
    CORE-POS.
    """)

    corepos_sku = sa.Column(sa.String(length=13), nullable=False, doc="""
    ``vendorItems.sku`` value for the corresponding record within
    CORE-POS.
    """)

    corepos_id = sa.Column(sa.Integer(), nullable=True, doc="""
    ``vendorItems.vendorItemID`` value for the corresponding record
    within CORE-POS.
    """)

    def __str__(self):
        return str(self.cost)

CoreProductCost.make_proxy(model.ProductCost, '_corepos', 'corepos_vendor_id')
CoreProductCost.make_proxy(model.ProductCost, '_corepos', 'corepos_sku')
CoreProductCost.make_proxy(model.ProductCost, '_corepos', 'corepos_id')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE POS (DB) -> Rattail data importing
"""

import datetime
import decimal
import logging
from collections import OrderedDict

import sqlalchemy as sa
from sqlalchemy import orm

from rattail import importing
from rattail.db.util import normalize_full_name
from rattail_corepos import importing as corepos_importing


log = logging.getLogger(__name__)


class FromCOREPOSToRattail(importing.FromSQLAlchemyHandler, importing.ToRattailHandler):
    """
    Import handler for data coming from a CORE POS database.
    """
    # TODO: these should be changed, it now allows for "trans" DB too..
    generic_host_title = 'CORE Office (DB "op")'
    host_key = 'corepos_db_office_op'
    corepos_dbkey = 'default'

    @property
    def host_title(self):
        return "CORE-POS (DB/{})".format(self.corepos_dbkey)

    def make_host_session(self):
        corepos = self.app.get_corepos_handler()

        # session type depends on the --corepos-dbtype arg
        if self.corepos_dbtype == 'office_trans':
            return corepos.make_session_office_trans(
                bind=self.config.coretrans_engines[self.corepos_dbkey])

        # assume office_op by default
        return corepos.make_session_office_op(
            bind=self.config.corepos_engines[self.corepos_dbkey])

    def get_importers(self):
        importers = OrderedDict()
        importers['Store'] = StoreImporter
        importers['Employee'] = EmployeeImporter
        importers['Customer'] = CustomerImporter
        importers['Member'] = MemberImporter
        importers['Tax'] = TaxImporter
        importers['Tender'] = TenderImporter
        importers['Vendor'] = VendorImporter
        importers['Department'] = DepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Product'] = ProductImporter
        importers['ProductCost'] = ProductCostImporter
        importers['MemberEquityPayment'] = MemberEquityPaymentImporter
        return importers

    def get_default_keys(self):
        keys = super().get_default_keys()

        if 'MemberEquityPayment' in keys:
            keys.remove('MemberEquityPayment')

        return keys


class FromCOREPOS(importing.FromSQLAlchemy):
    """
    Base class for all CORE POS data importers.
    """

    def setup(self):
        super().setup()
        self.ignore_new_members = self.should_ignore_new_members()

    def should_ignore_new_members(self):
        if hasattr(self, 'ignore_new_members'):
            return self.ignore_new_members

        return self.config.getbool('rattail_corepos',
                                   'importing_ignore_new_members',
                                   default=False)

    def is_new_member(self, member):
        if member.customers:
            customer = member.customers[0]
            if customer.last_name == 'NEW MEMBER'and not customer.first_name:
                return True
        return False


class StoreImporter(FromCOREPOS, corepos_importing.model.StoreImporter):
    """
    Importer for store data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Store

    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'id',
        'name',
    ]

    def normalize_host_object(self, store):
        return {
            'corepos_id': store.id,
            'id': str(store.id),
            'name': store.description,
        }


class EmployeeImporter(FromCOREPOS, corepos_importing.model.EmployeeImporter):
    """
    Importer for employee data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Employee

    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'first_name',
        'last_name',
        'full_name',
        'status',
    ]

    def normalize_host_object(self, employee):
        return {
            'corepos_number': employee.number,
            'first_name': employee.first_name,
            'last_name': employee.last_name,
            'full_name': normalize_full_name(employee.first_name, employee.last_name),
            'status': self.enum.EMPLOYEE_STATUS_CURRENT if employee.active else self.enum.EMPLOYEE_STATUS_FORMER,
        }


class CustomerImporter(FromCOREPOS, corepos_importing.model.CustomerImporter):
    """
    Importer for customer data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.MemberInfo

    key = 'corepos_card_number'
    supported_fields = [
        'corepos_card_number',
        'number',
        'name',
        'account_holder_uuid',
        'account_holder_first_name',
        'account_holder_last_name',
        'account_holder_full_name',
        'email_address',
        'phone_number',
        'address_street',
        'address_street2',
        'address_city',
        'address_state',
        'address_zipcode',
    ]

    def setup(self):
        super().setup()
        model = self.model

        query = self.session.query(model.Person)\
                            .outerjoin(model.Customer,
                                       model.Customer.account_holder_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreCustomer)\
                            .outerjoin(model.Member,
                                       model.Member.person_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreMember)\
                            .filter(sa.or_(
                                model.CoreCustomer.corepos_card_number != None,
                                model.CoreMember.corepos_card_number != None))\
                            .options(orm.joinedload(model.Person.customer_accounts)\
                                     .joinedload(model.Customer._corepos))
        def card_number(person, normal):
            customer = self.app.get_customer(person)
            if customer and customer.corepos_card_number:
                return customer.corepos_card_number
            member = self.app.get_member(person)
            if member and member.corepos_card_number:
                return member.corepos_card_number
        self.people_by_card_number = self.cache_model(model.Person, query=query,
                                                      key=card_number)

    def get_person(self, member):

        if hasattr(self, 'people_by_card_number'):
            return self.people_by_card_number.get(member.card_number)

        model = self.model

        try:
            return self.session.query(model.Person)\
                               .join(model.Customer,
                                     model.Customer.account_holder_uuid == model.Person.uuid)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == member.card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

        try:
            return self.session.query(model.Person)\
                               .join(model.Member,
                                     model.Member.person_uuid == model.Person.uuid)\
                               .join(model.CoreMember)\
                               .filter(model.CoreMember.corepos_card_number == member.card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, member):
        card_number = member.card_number

        # maybe ignore NEW MEMBER accounts
        if self.should_ignore_new_members():
            if self.is_new_member(member):
                return

        contact = member
        if member.customers:
            contact = member.customers[0]

        person = self.get_person(member)

        street = (member.street or '').split('\n')

        full_name = normalize_full_name(contact.first_name, contact.last_name)
        return {
            'corepos_card_number': card_number,
            'number': card_number,
            'name': full_name,
            'account_holder_uuid': person.uuid if person else None,
            'account_holder_first_name': contact.first_name,
            'account_holder_last_name': contact.last_name,
            'account_holder_full_name': full_name,
            'email_address': (member.email or '').strip() or None,
            'phone_number': self.app.format_phone_number((member.phone or '').strip() or None),
            'address_street': street[0] or None,
            'address_street2': (street[1] or None) if len(street) > 1 else None,
            'address_city': member.city or None,
            'address_state': member.state or None,
            'address_zipcode': member.zip or None,
        }


class MemberImporter(FromCOREPOS, corepos_importing.model.MemberImporter):
    """
    Importer for member data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.MemberInfo

    # TODO use this key instead
    #key = 'corepos_card_number'
    key = 'number'
    supported_fields = [
        'number',
        'corepos_card_number',
        'customer_uuid',
        'person_uuid',
        'person_first_name',
        'person_last_name',
        'membership_type_number',
        'joined',
        'withdrew',
        'active',
    ]

    def setup(self):
        super().setup()
        model = self.model

        self.customers_by_number = self.app.cache_model(self.session,
                                                        model.Customer,
                                                        key='number')

        query = self.session.query(model.Person)\
                            .outerjoin(model.Customer,
                                       model.Customer.account_holder_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreCustomer)\
                            .outerjoin(model.Member,
                                       model.Member.person_uuid == model.Person.uuid)\
                            .outerjoin(model.CoreMember)\
                            .filter(sa.or_(
                                model.CoreCustomer.corepos_card_number != None,
                                model.CoreMember.corepos_card_number != None))\
                            .options(orm.joinedload(model.Person.customer_accounts)\
                                     .joinedload(model.Customer._corepos))
        def card_number(person, normal):
            customer = self.app.get_customer(person)
            if customer and customer.corepos_card_number:
                return customer.corepos_card_number
            member = self.app.get_member(person)
            if member and member.corepos_card_number:
                return member.corepos_card_number
        self.people_by_card_number = self.cache_model(model.Person, query=query,
                                                      key=card_number)

    def get_person(self, member):
        if hasattr(self, 'people_by_card_number'):
            return self.people_by_card_number.get(member.card_number)

        model = self.model

        try:
            return self.session.query(model.Person)\
                               .join(model.Customer,
                                     model.Customer.account_holder_uuid == model.Person.uuid)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == member.card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

        try:
            return self.session.query(model.Person)\
                               .join(model.Member,
                                     model.Member.person_uuid == model.Person.uuid)\
                               .join(model.CoreMember)\
                               .filter(model.CoreMember.corepos_card_number == member.card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_customer_by_number(self, number):
        if hasattr(self, 'customers_by_number'):
            return self.customers_by_number.get(number)

        model = self.model
        try:
            return self.session.query(model.Customer)\
                               .filter(model.Customer.number == number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, core_member):

        # maybe ignore NEW MEMBER accounts
        if self.should_ignore_new_members():
            if self.is_new_member(core_member):
                return

        core_customer = core_member.customers[0] if core_member.customers else None
        core_contact = core_customer or core_member

        card_number = core_member.card_number
        customer = self.get_customer_by_number(card_number)
        person = self.get_person(core_member)

        typeno = None
        if core_customer and core_customer.member_type:
            typeno = core_customer.member_type.id

        joined = None
        withdrew = None
        if core_member.dates:
            dates = core_member.dates
            joined = dates.start_date.date() if dates.start_date else None
            withdrew = dates.end_date.date() if dates.end_date else None
            if joined and joined == datetime.date(1900, 1, 1):
                joined = None
            if withdrew and withdrew == datetime.date(1900, 1, 1):
                withdrew = None

        return {
            'number': card_number,
            'corepos_card_number': card_number,
            'customer_uuid': customer.uuid if customer else None,
            'person_uuid': person.uuid if person else None,
            'person_first_name': core_contact.first_name,
            'person_last_name': core_contact.last_name,
            'membership_type_number': typeno,
            'joined': joined,
            'withdrew': withdrew,
            'active': not bool(withdrew),
        }


class TaxImporter(FromCOREPOS, corepos_importing.model.TaxImporter):
    """
    Importer for tax data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.TaxRate

    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'code',
        'description',
        'rate',
    ]

    def normalize_host_object(self, tax):
        return {
            'corepos_id': tax.id,
            'code': str(tax.id),
            'description': tax.description,
            'rate': decimal.Decimal(str(tax.rate * 100)),
        }


class TenderImporter(FromCOREPOS, corepos_importing.model.TenderImporter):
    """
    Importer for tender data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Tender

    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'code',
        'name',
    ]

    def normalize_host_object(self, tender):
        return {
            'corepos_id': tender.tender_id,
            'code': tender.tender_code,
            'name': tender.tender_name,
        }


class VendorImporter(FromCOREPOS, corepos_importing.model.VendorImporter):
    """
    Importer for vendor data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Vendor

    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'name',
        'abbreviation',
        'special_discount',
        'phone_number',
        'fax_number',
        'email_address',
    ]

    def cache_query(self):
        """
        Return the query to be used when caching "local" data.
        """
        # can't just use rattail.db.model b/c the CoreVendor would normally not
        # be in there!  this still requires custom model to be configured though.
        model = self.app.model

        # first get default query
        query = super().cache_query()

        # maybe filter a bit, to ensure only "relevant" records are involved
        if 'corepos_id' in self.key:
            # note, the filter is probably redundant since we INNER JOIN on the
            # extension table, and it doesn't allow null ID values.  but clarity.
            query = query.join(model.CoreVendor)\
                         .filter(model.CoreVendor.corepos_id != None)

        return query

    def normalize_host_object(self, vendor):

        special_discount = None
        if vendor.discount_rate is not None:
            special_discount = decimal.Decimal('{:0.3f}'.format(vendor.discount_rate))

        return {
            'corepos_id': vendor.id,
            'name': vendor.name,
            'abbreviation': vendor.abbreviation or None,
            'special_discount': special_discount,
            'phone_number': vendor.phone or None,
            'fax_number': vendor.fax or None,
            'email_address': vendor.email or None,
        }


class DepartmentImporter(FromCOREPOS, corepos_importing.model.DepartmentImporter):
    """
    Importer for department data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Department

    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'tax_code',
        'food_stampable',
    ]

    def normalize_host_object(self, department):
        return {
            'corepos_number': department.number,
            'number': department.number,
            'name': department.name,
            'tax_code': str(department.tax_rate.id if department.tax_rate else '') or None,
            'food_stampable': department.food_stampable,
        }


class SubdepartmentImporter(FromCOREPOS, corepos_importing.model.SubdepartmentImporter):
    """
    Importer for subdepartment data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Subdepartment

    key = 'corepos_number'
    supported_fields = [
        'corepos_number',
        'number',
        'name',
        'department_number',
    ]

    def normalize_host_object(self, subdepartment):
        return {
            'corepos_number': subdepartment.number,
            'number': subdepartment.number,
            'name': subdepartment.name,
            'department_number': subdepartment.department_number,
        }


class ProductImporter(FromCOREPOS, corepos_importing.model.ProductImporter):
    """
    Importer for product data from CORE POS.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.Product

    key = 'corepos_id'
    supported_fields = [
        'corepos_id',
        'item_id',
        'upc',
        'brand_name',
        'description',
        'size',
        'unit_size',
        'unit_of_measure',
        'uom_abbreviation',
        'weighed',
        'department_number',
        'subdepartment_number',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_type',
        'sale_price_price',
        'sale_price_starts',
        'sale_price_ends',
        'sale_price_current',
        'food_stampable',
        'discountable',
        # 'tax1',
        'tax_code',
        'not_for_sale',
    ]

    def setup(self):
        super().setup()
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()

        if self.fields_active(self.sale_price_fields):
            self.core_batch_items = {}

            # TODO: it seems possible for CORE to have more than one
            # batch item for a given product.  sort order will
            # determine which would "win" but not clear what sort
            # order should be used, e.g. CORE does not seem to use one
            today = self.app.today()
            batches = self.host_session.query(op_model.Batch)\
                                       .filter(op_model.Batch.start_date <= today)\
                                       .filter(op_model.Batch.end_date >= today)\
                                       .filter(op_model.Batch.discount_type > 0)\
                                       .options(orm.joinedload(op_model.Batch.items))\
                                       .all()

            def cache(batch, i):
                for item in batch.items:
                    self.core_batch_items.setdefault(item.upc, []).append(item)

            self.progress_loop(cache, batches,
                               message="Caching CORE-POS batch items")

    def get_core_batch_item(self, upc):
        if hasattr(self, 'core_batch_items'):
            items = self.core_batch_items.get(upc)
            if not items:
                return

            sale_price = items[0].sale_price
            if any([item.sale_price != sale_price
                    for item in items[1:]]):
                log.warning("ambiguous batch items for upc: %s", upc)

            return items[0]

        raise NotImplementedError("TODO: fetch batch items in real-time")

    def normalize_host_object(self, product):

        try:
            upc = self.app.make_gpc(product.upc, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.debug("CORE POS product has invalid UPC: %s", product.upc)
            if len(self.key) == 1 and self.key[0] == 'upc':
                return
            upc = None

        price = None
        if product.normal_price is not None:
            price = decimal.Decimal('{:03f}'.format(product.normal_price))

        data = {
            'corepos_id': product.id,
            'item_id': product.upc,
            'upc': upc,
            'brand_name': (product.brand or '').strip() or None,
            'description': (product.description or '').strip(),

            'department_number': product.department_number or None,
            'subdepartment_number': product.subdepartment_number or None,

            'weighed': bool(product.scale),
            'food_stampable': product.foodstamp,
            # 'tax1': bool(product.tax), # TODO: is this right?

            'regular_price_price': price,
            'regular_price_multiple': 1 if price is not None else None,
            'regular_price_type': self.enum.PRICE_TYPE_REGULAR if price is not None else None,

            # nb. these may get set below
            'sale_price_price': None,
            'sale_price_starts': None,
            'sale_price_ends': None,
            'sale_price_current': False,

            'discountable': bool(product.line_item_discountable),

            'not_for_sale': not product.in_use,
        }

        if 'tax_code' in self.fields:
            data['tax_code'] = str(product.tax_rate.id) if product.tax_rate else None

        if self.fields_active(self.size_fields):
            size_info = self.normalize_size_info(product)
            data.update({
                'size': size_info['size'],
                'unit_size': size_info['unit_size'],
                'unit_of_measure': size_info['uom_code'],
                'uom_abbreviation': size_info['uom_abbrev'],
            })

        # sale price
        # nb. CORE discount_type indicates if item is on sale "now"
        if self.fields_active(self.sale_price_fields) and product.discount_type:
            item = self.get_core_batch_item(product.upc)
            if item:
                data.update({
                    'sale_price_price': item.sale_price,
                    'sale_price_starts': self.app.make_utc(self.app.localtime(item.batch.start_date)),
                    'sale_price_ends': self.app.make_utc(self.app.localtime(item.batch.end_date)),
                    'sale_price_current': True,
                })

        return data


class ProductCostImporter(FromCOREPOS, corepos_importing.model.ProductCostImporter):
    """
    Importer for product cost data from CORE POS API.
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        op_model = corepos.get_model_office_op()
        return op_model.VendorItem

    key = ('corepos_vendor_id', 'corepos_sku')
    supported_fields = [
        'corepos_vendor_id',
        'corepos_sku',
        'product_uuid',
        'vendor_uuid',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preferred',
    ]

    def setup(self):
        super().setup()
        model = self.model

        query = self.session.query(model.Vendor)\
                            .join(model.CoreVendor)\
                            .filter(model.CoreVendor.corepos_id != None)
        self.vendors_by_corepos_id = self.cache_model(model.Vendor,
                                                      query=query,
                                                      key='corepos_id')

        self.products_by_item_id = self.cache_model(model.Product, key='item_id')

    def query(self):
        query = super().query()

        query = query.options(orm.joinedload(corepos.VendorItem.product))

        return query

    def get_vendor(self, item):
        corepos_id = item.vendor_id

        if hasattr(self, 'vendors_by_corepos_id'):
            return self.vendors_by_corepos_id.get(corepos_id)

        model = self.app.model
        try:
            return self.session.query(model.Vendor)\
                               .join(model.CoreVendor)\
                               .filter(model.CoreVendor.corepos_id == corepos_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_product(self, item):
        item_id = item.upc

        if hasattr(self, 'products_by_item_id'):
            return self.products_by_item_id.get(item_id)

        model = self.model
        try:
            return self.session.query(model.Product)\
                               .filter(model.Product.item_id == item_id)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def normalize_host_object(self, item):

        vendor = self.get_vendor(item)
        if not vendor:
            log.warning("CORE POS vendor not found for item: %s", item)
            return

        product = self.get_product(item)
        if not product:
            # just debug logging since this is a common scenario; the
            # CORE table is for items "available from vendor" but not
            # necssarily items carried by store
            log.debug("product not found for CORE vendor item: %s", item)
            return

        core_product = item.product

        preferred = False
        if core_product and core_product.default_vendor_id == item.vendor_id:
            preferred = True

        case_size = decimal.Decimal(str(item.units))
        unit_cost = item.cost
        case_cost = None
        if unit_cost is not None:
            case_cost = unit_cost * case_size

        return {
            'corepos_vendor_id': item.vendor_id,
            'corepos_sku': item.sku,
            'product_uuid': product.uuid,
            'vendor_uuid': vendor.uuid,
            'code': item.sku,
            'case_size': case_size,
            'case_cost': case_cost,
            'unit_cost': unit_cost,
            'preferred': preferred,
        }


class MemberEquityPaymentImporter(FromCOREPOS, corepos_importing.model.MemberEquityPaymentImporter):
    """
    Imports equity payment data from CORE-POS
    """

    @property
    def host_model_class(self):
        corepos = self.app.get_corepos_handler()
        trans_model = corepos.get_model_office_trans()
        return trans_model.StockPurchase

    key = 'uuid'
    supported_fields = [
        'uuid',
        'member_uuid',
        'amount',
        'received',
        'transaction_identifier',
        'corepos_card_number',
        'corepos_transaction_number',
        'corepos_transaction_id',
        'corepos_department_number',
        'corepos_datetime',
    ]

    def setup(self):
        super().setup()
        model = self.model

        query = self.session.query(model.Member)\
                            .join(model.Customer)\
                            .join(model.CoreCustomer)\
                            .options(orm.joinedload(model.Member.customer)\
                                     .joinedload(model.Customer._corepos))
        self.members_by_card_number = self.cache_model(
            model.Member, query=query,
            key=lambda member, normal: member.customer.corepos_card_number)

        payments = self.session.query(model.MemberEquityPayment)\
                               .join(model.CoreMemberEquityPayment)\
                               .filter(model.CoreMemberEquityPayment.corepos_transaction_number != None)\
                               .filter(model.CoreMemberEquityPayment.corepos_transaction_id != None)\
                               .all()
        self.payments_by_card_number = {}
        def add(payment, i):
            self.payments_by_card_number.setdefault(payment.corepos_card_number, []).append(payment)
        self.progress_loop(add, payments, message="Grouping payments by card number")

    def cache_query(self):
        query = super().cache_query()
        model = self.model

        # ignore existing payments not known to be in CORE
        # TODO: is this in fact a good idea?
        query = query.join(model.CoreMemberEquityPayment)\
                     .filter(model.CoreMemberEquityPayment.corepos_transaction_number != None)

        return query

    def get_member(self, card_number):
        if hasattr(self, 'members_by_card_number'):
            return self.members_by_card_number.get(card_number)

        model = self.model
        try:
            return self.session.query(model.Member)\
                               .join(model.Customer)\
                               .join(model.CoreCustomer)\
                               .filter(model.CoreCustomer.corepos_card_number == card_number)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def get_payment(self, stock_purchase):
        payments = self.payments_by_card_number.get(stock_purchase.card_number)
        if not payments:
            return

        # first look for exact match
        dt = self.app.localtime(stock_purchase.datetime)
        match = [payment for payment in payments
                 if payment.corepos_transaction_number == stock_purchase.transaction_number
                 and payment.corepos_transaction_id == stock_purchase.transaction_id
                 and payment.amount == stock_purchase.amount
                 and payment.corepos_department_number == stock_purchase.department_number
                 and self.app.localtime(payment.corepos_datetime, from_utc=True) == dt]
        if len(match) == 1:
            return match[0]

        # nb. avoid datetime for this one
        matches = [payment for payment in payments
                   if payment.corepos_transaction_number == stock_purchase.transaction_number
                   and payment.corepos_transaction_id == stock_purchase.transaction_id
                   and payment.amount == stock_purchase.amount
                   and payment.corepos_department_number == stock_purchase.department_number]
        if len(matches) == 1:
            log.warning("found 'loose' match for card #%s, txn %s, for $%0.2f: %s",
                        stock_purchase.card_number,
                        stock_purchase.transaction_number,
                        stock_purchase.amount,
                        stock_purchase.datetime)

            # so there is one match, but its timestamp may be way off,
            # so let's also make sure at least date matches
            payment = matches[0]
            if self.app.localtime(payment.corepos_datetime, from_utc=True).date() == dt.date():
                return payment

            # do not assume any match, if dates were off
            return

        # TODO: not sure how to handle yet, if last check found more
        # than one match.  presumably if none were found then it is
        # safe to just return none though..
        if matches:
            raise NotImplementedError(f"{len(matches)} payments matched for CORE equity: {stock_purchase}")

    def normalize_host_object(self, stock_purchase):

        card_number = stock_purchase.card_number
        member = self.get_member(card_number)
        if not member:
            log.warning("member not found for card number %s: %s",
                        card_number, stock_purchase)
            return

        dt = stock_purchase.datetime
        if dt:
            dt = self.app.make_utc(self.app.localtime(dt))
        corepos_datetime = dt
        received = dt

        payment = self.get_payment(stock_purchase)
        if payment and payment.source != 'corepos':
            received = payment.received

        return {
            'uuid': payment.uuid if payment else self.app.make_uuid(),
            'member_uuid': member.uuid,
            'amount': stock_purchase.amount,
            'received': received,
            'transaction_identifier': stock_purchase.transaction_number,
            'corepos_card_number': stock_purchase.card_number,
            'corepos_transaction_number': stock_purchase.transaction_number,
            'corepos_transaction_id': stock_purchase.transaction_id,
            'corepos_department_number': stock_purchase.department_number,
            'corepos_datetime': corepos_datetime,
        }

    def create_object(self, key, host_data):
        payment = super().create_object(key, host_data)
        if payment:

            # track where each payment comes from!
            payment.source = 'corepos'

            return payment



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office - patch customer gaps
"""

import logging

from rattail.app import GenericHandler
from rattail_corepos.corepos.util import get_core_members


log = logging.getLogger(__name__)


class CustomerGapPatcher(GenericHandler):
    """
    POST to the CORE API as needed, to patch gaps for customerID
    """

    def run(self, dry_run=False, progress=None):
        corepos = self.app.get_corepos_handler()
        op = corepos.get_model_office_op()
        corepos_api = corepos.make_webapi()
        members = get_core_members(self.config, corepos_api, progress=progress)
        tally = self.app.make_object(updated=0)

        self.maxlen_phone = self.app.maxlen(op.Customer.phone)
        # nb. just in case the smallest one changes in future..
        other = self.app.maxlen(op.MemberInfo.phone)
        if other < self.maxlen_phone:
            self.maxlen_phone = other

        def inspect(member, i):
            for customer in member['customers']:
                customer_id = int(customer['customerID'])
                if not customer_id:
                    data = dict(member)
                    self.trim_phones(data)
                    cardno = data.pop('cardNo')
                    log.debug("%s call set_member() for card no %s: %s",
                              'should' if dry_run else 'will',
                              cardno, data)
                    if not dry_run:
                        corepos_api.set_member(cardno, **data)
                    tally.updated += 1
                    return

        action = "Finding"
        if not dry_run:
            action += " and fixing"
        self.app.progress_loop(inspect, members, progress,
                               message=f"{action} customerID gaps")

        sys.stdout.write("\n")
        if dry_run:
            sys.stdout.write("would have ")
        sys.stdout.write(f"updated {tally.updated} members\n")

    def trim_phones(self, data):
        # the `meminfo` table allows 30 chars for phone, but
        # `Customers` table only allows 20 chars.  so we must trim to
        # 20 chars or else the CORE API will silently fail to update
        # tables correctly when we POST to it
        for customer in data['customers']:
            for field in ['phone', 'altPhone']:
                value = customer[field]
                if len(value) > self.maxlen_phone:
                    log.warning("phone value for cardno %s is too long (%s chars) "
                                "and will be trimmed to %s chars: %s",
                                data['cardNo'],
                                len(value),
                                self.maxlen_phone,
                                value)
                    customer[field] = value[:self.maxlen_phone]



# -*- coding: utf-8; -*-
"""add corepos_tax

Revision ID: 673ff7088d35
Revises: 1f2e2f57c90b
Create Date: 2023-10-06 21:01:01.105790

"""

# revision identifiers, used by Alembic.
revision = '673ff7088d35'
down_revision = '1f2e2f57c90b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_tax
    op.create_table('corepos_tax',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['uuid'], ['tax.uuid'], name='corepos_tax_fk_tax'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('corepos_tax_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_corepos_tax_version_end_transaction_id'), 'corepos_tax_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_tax_version_operation_type'), 'corepos_tax_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_tax_version_transaction_id'), 'corepos_tax_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_tax
    op.drop_index(op.f('ix_corepos_tax_version_transaction_id'), table_name='corepos_tax_version')
    op.drop_index(op.f('ix_corepos_tax_version_operation_type'), table_name='corepos_tax_version')
    op.drop_index(op.f('ix_corepos_tax_version_end_transaction_id'), table_name='corepos_tax_version')
    op.drop_table('corepos_tax_version')
    op.drop_table('corepos_tax')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
POS batch handler, for CORE-POS integration
"""

import logging

from rattail.batch import pos as base


log = logging.getLogger(__name__)


class POSBatchHandler(base.POSBatchHandler):
    """
    Handler for POS batches
    """

    def describe_execution(self, batch, **kwargs):
        return ("A new transaction will be created in CORE Office, directly "
                "in `dtransactions`, to mirror this batch.")

    def execute(self, batch, progress=None, **kwargs):
        rows = self.get_effective_rows(batch)
        if not rows:
            return True

        self.corepos_handler = self.app.get_corepos_handler()
        self.coretrans = self.corepos_handler.get_model_office_trans()
        self.maxlen_upc = self.app.maxlen(self.coretrans.TransactionDetail.upc)
        self.maxlen_description = self.app.maxlen(self.coretrans.TransactionDetail.description)

        # convert batch rows to `dtransactions` records
        dtransactions = self.normalize_dtransactions(batch, rows, progress)
        if not dtransactions:
            return True

        # commit all to `dtransactions`
        coretrans_session = self.corepos_handler.make_session_office_trans()
        coretrans_session.add_all(dtransactions)
        coretrans_session.commit()
        coretrans_session.close()
        return True

    def normalize_dtransactions(self, batch, rows, progress=None):
        dtransactions = []
        self.made_subtotal = False

        def add(row, i):

            if row.row_type == self.enum.POS_ROW_TYPE_SELL:
                d = self.make_d_item(row)
                dtransactions.append(d)

            elif row.row_type == self.enum.POS_ROW_TYPE_BADSCAN:
                d = self.make_d_badscan(row)
                dtransactions.append(d)

            elif row.row_type in (self.enum.POS_ROW_TYPE_SET_CUSTOMER,
                                  self.enum.POS_ROW_TYPE_SWAP_CUSTOMER):
                d = self.make_d_customer(row)
                dtransactions.append(d)

            elif row.row_type == self.enum.POS_ROW_TYPE_TENDER:

                if not self.made_subtotal:
                    d = self.make_d_subtotal(row, dtransactions)
                    dtransactions.append(d)
                    self.made_subtotal = True

                d = self.make_d_tender(row)
                dtransactions.append(d)

            elif row.row_type == self.enum.POS_ROW_TYPE_CHANGE_BACK:

                d = self.make_d_change(row)
                dtransactions.append(d)

                d = self.make_d_discount(batch, dtransactions)
                dtransactions.append(d)

                d = self.make_d_tax(batch, dtransactions)
                dtransactions.append(d)

        self.progress_loop(add, rows, progress,
                           message="Normalizing items for CORE-POS transaction")

        # now that we have all records, fill in some more values
        session = self.app.get_session(batch)
        store = self.config.get_store(session)
        store_id = store.corepos_id if store else None
        register_number = int(batch.terminal_id)
        employee_number = batch.cashier.corepos_number
        member = self.app.get_member(batch.customer)
        member_type = member.membership_type.number if member else None
        pos_row_id = f'corepos_pos_row_id_term_{batch.terminal_id}'
        self.app.make_counter(session, pos_row_id)
        for i, d in enumerate(dtransactions, 1):
            d.store_id = store_id
            d.register_number = register_number
            d.employee_number = employee_number
            d.card_number = batch.customer.number
            d.member_type = member_type
            d.staff = batch.customer_is_employee
            d.transaction_id = i
            d.pos_row_id = self.app.next_counter_value(session, pos_row_id)

        return dtransactions

    def make_d_basic(self, batch=None, row=None):
        if not batch and not row:
            raise ValueError("must specify either batch or row")
        
        if not batch:
            batch = row.batch

        d = self.coretrans.TransactionDetail()

        d.transaction_number = batch.id

        if row and row.timestamp:
            d.date_time = self.app.localtime(row.timestamp, from_utc=True)
        else:
            # nb. batch.created *should* have a value..if not this would be "now"
            d.date_time = self.app.localtime(batch.created, from_utc=True)

        # TODO: i *think* all these are safe defaults, and can
        # override per line item as needed
        d.transaction_status = ''
        d.department_number = 0
        d.unit_price = 0
        d.reg_price = 0
        d.tax_rate_id = 0
        d.food_stamp = False
        d.member_discount = 0
        d.discount_type = 0
        d.percent_discount = 0
        d.quantity = 0
        d.item_quantity = 0
        d.volume_discount_type = 0
        d.volume_special = 0
        d.mix_match = 0
        d.upc = '0'
        d.num_flag = 0
        d.char_flag = ''
        d.cost = 0
        d.discount = 0
        d.discountable = False
        d.total = 0
        d.voided = 0
        d.volume = 0
        d.matched = False

        return d

    def make_d_badscan(self, row):
        d = self.make_d_basic(row=row)

        d.description = 'BADSCAN'

        d.upc = row.item_entry
        if d.upc and len(d.upc) > self.maxlen_upc:
            log.debug("have to truncate this upc to %s chars (it has %s): %s",
                      self.maxlen_upc, len(d.upc), d.upc)
            d.upc = d.upc[:self.maxlen_upc]
            d.description += " (TRUNCATED)"

        return d

    def make_d_customer(self, row):
        batch = row.batch
        d = self.make_d_basic(row=row)

        d.upc = 'MEMENTRY'
        d.description = 'CARDNO IN NUMFLAG'

        # TODO: what do these mean? are they correct?
        d.transaction_type = 'L'
        d.transaction_subtype = 'OG'
        d.transaction_status = 'D'
        d.num_flag = batch.customer.number
        d.char_flag = '1'

        return d

    def make_d_item(self, row):
        batch = row.batch
        session = self.app.get_session(batch)
        d = self.make_d_basic(batch, row)

        d.transaction_type = 'I'
        d.transaction_subtype = 'NA'
        d.upc = row.product.item_id
        d.department_number = row.department_number
        d.food_stamp = row.foodstamp_eligible

        d.description = row.product.description
        if d.description and len(d.description) > self.maxlen_description:
            log.debug("have to truncate this description to %s chars (it has %s): %s",
                      self.maxlen_description, len(d.description), d.description)
            d.description = d.description[:self.maxlen_description]

        # TODO: should item_quantity ever differ?  see also
        # https://github.com/CORE-POS/IS4C/wiki/Office-Transaction-Database#dtransactions
        d.quantity = row.quantity
        d.item_quantity = row.quantity

        if row.product.cost:
            d.cost = row.product.cost.unit_cost

        d.unit_price = row.txn_price
        d.reg_price = row.reg_price
        d.discountable = row.product.discountable

        d.tax_rate_id = 0
        if row.tax_code:
            tax = self.get_tax(session, row.tax_code)
            d.tax_rate_id = tax.corepos_id
            if not d.tax_rate_id:
                log.error("tax not found in CORE-POS: %s", row.tax_code)
                d.tax_rate_id = 0

        d.total = row.sales_total
        # TODO: if void, should the above change too?
        d.voided = 1 if row.void else 0

        return d

    def make_d_subtotal(self, row, dtransactions):
        batch = row.batch
        d = self.make_d_basic(batch, row)

        d.transaction_type = 'C'
        d.transaction_status = 'D'
        d.voided = 3            # TODO (?)

        d.unit_price = sum([detail.total
                            for detail in dtransactions])

        # TODO
        tax = 0

        d.description = f"Subtotal {d.unit_price:0.2f}, Tax{tax:0.2f} #{batch.customer.number}"

        return d

    def make_d_tender(self, row):
        batch = row.batch
        d = self.make_d_basic(batch, row)

        d.transaction_type = 'T'
        d.transaction_subtype = row.item_entry
        d.description = row.description
        d.total = row.tender_total

        return d

    def make_d_change(self, row):
        batch = row.batch
        d = self.make_d_basic(batch, row)

        d.transaction_type = 'T'
        d.transaction_subtype = row.item_entry
        d.description = "Change"
        d.total = row.tender_total

        if not d.total:
            d.voided = 8        # TODO (?)

        return d

    def make_d_discount(self, batch, dtransactions):
        d = self.make_d_basic(batch)

        d.transaction_type = 'S'
        d.upc = 'DISCOUNT'
        d.quantity = 1
        d.item_quantity = 1
        d.description = "Discount"

        return d

    def make_d_tax(self, batch, dtransactions):
        d = self.make_d_basic(batch)

        d.transaction_type = 'A'
        d.upc = 'TAX'
        d.description = "Tax"

        return d



# -*- coding: utf-8; -*-
"""add MemberEquityPayment extension

Revision ID: 08d879bbe118
Revises: b025df7cf41b
Create Date: 2023-09-06 17:44:43.874500

"""

# revision identifiers, used by Alembic.
revision = '08d879bbe118'
down_revision = 'b025df7cf41b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # corepos_member_equity_payment
    op.create_table('corepos_member_equity_payment',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('corepos_card_number', sa.Integer(), nullable=False),
                    sa.Column('corepos_transaction_number', sa.String(length=50), nullable=True),
                    sa.Column('corepos_transaction_id', sa.Integer(), nullable=True),
                    sa.Column('corepos_department_number', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['uuid'], ['member_equity_payment.uuid'], name='corepos_member_equity_payment_fk_payment'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('corepos_member_equity_payment_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('corepos_card_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('corepos_transaction_number', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('corepos_transaction_id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('corepos_department_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_corepos_member_equity_payment_version_end_transaction_id'), 'corepos_member_equity_payment_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_corepos_member_equity_payment_version_operation_type'), 'corepos_member_equity_payment_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_corepos_member_equity_payment_version_transaction_id'), 'corepos_member_equity_payment_version', ['transaction_id'], unique=False)


def downgrade():

    # corepos_member_equity_payment
    op.drop_index(op.f('ix_corepos_member_equity_payment_version_transaction_id'), table_name='corepos_member_equity_payment_version')
    op.drop_index(op.f('ix_corepos_member_equity_payment_version_operation_type'), table_name='corepos_member_equity_payment_version')
    op.drop_index(op.f('ix_corepos_member_equity_payment_version_end_transaction_id'), table_name='corepos_member_equity_payment_version')
    op.drop_table('corepos_member_equity_payment_version')
    op.drop_table('corepos_member_equity_payment')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Handler for use with CORE-POS
"""

from sqlalchemy import orm

from rattail import auth as base


class CoreAuthHandler(base.AuthHandler):
    """
    Custom auth handler for use with CORE-POS.
    """

    def authenticate_user(self, session, username, password):
        model = self.model

        # first try default logic, if it works then great
        user = super().authenticate_user(session, username, password)
        if user:
            return user

        # only if configured, we also check CORE-POS credentials
        if self.config.getbool('rattail.auth', 'corepos.check_cashier_credentials',
                               default=False):
            user = None
            core_session = self.app.get_corepos_handler().make_session_office_op()
            core_employee = self.check_corepos_cashier_credentials(core_session, password)
            if core_employee:
                user = self.get_user_from_corepos_employee(session, core_employee)
                if not user and self.config.get_bool('rattail.auth.corepos.automake_users'):
                    # nb. new user must be made via separate session
                    # and then merged back into the main session.
                    # this is because the caller cannot be responsible
                    # for committing (persisting) the new user.
                    with self.app.short_session() as s:
                        user = self.make_user_from_corepos_employee(s, core_employee)
                        s.commit()
                        user = session.get(model.User, user.uuid)
            core_session.close()
            if user and user.active:
                return user

    def check_corepos_cashier_credentials(self, core_session, password):
        core_op = self.app.get_corepos_handler().get_model_office_op()

        try:
            core_employee = core_session.query(core_op.Employee)\
                                        .filter(core_op.Employee.cashier_password == password)\
                                        .one()
        except orm.exc.NoResultFound:
            pass
        else:
            if core_employee.active:
                return core_employee

    def get_rattail_employee(self, session, core_employee):
        model = self.model
        try:
            return session.query(model.Employee)\
                          .join(model.CoreEmployee)\
                          .filter(model.CoreEmployee.corepos_number == core_employee.number)\
                          .one()
        except orm.exc.NoResultFound:
            pass

    def get_user_from_corepos_employee(self, session, core_employee):
        employee = self.get_rattail_employee(session, core_employee)
        if employee:
            return self.app.get_user(employee)

    def make_user_from_corepos_employee(self, session, core_employee):
        employee = self.get_rattail_employee(session, core_employee)
        if not employee:
            raise ValueError(f"CORE employee not found in {self.app.get_title()}: {core_employee}")

        person = self.app.get_person(employee)
        user = self.make_user(session=session, person=person)

        role = self.config.get('rattail.auth.corepos.automake_users_role')
        if role:
            role = self.get_role(session, role)
            if role:
                user.roles.append(role)

        return user



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV -> CORE data import
"""

from rattail.importing.handlers import FromFileHandler
from rattail.importing.csv import FromCSVToSQLAlchemyMixin
from rattail_corepos.corepos.office.importing.db.model import ToCoreHandler, ToCore


class FromCSVToCore(FromCSVToSQLAlchemyMixin, FromFileHandler, ToCoreHandler):
    """
    Handler for CSV -> CORE data import
    """
    host_title = "CSV"
    ToParent = ToCore

    def get_model(self):

        if self.corepos_dbtype == 'office_op':
            from corepos.db.office_op import model
            return model

        if self.corepos_dbtype == 'office_trans':
            from corepos.db.office_trans import model
            return model

        if self.corepos_dbtype == 'office_arch':
            from corepos.db.office_arch import model
            return model

        raise ValueError(f"unknown corepos_dbtype: {self.corepos_dbtype}")

    def get_default_keys(self):
        keys = super().get_default_keys()

        # error will occur for any missing files, when running for all
        # default models.  so make sure some of these are not default.
        if 'Change' in keys:
            keys.remove('Change')

        return keys



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database schema extensions for CORE-POS integration
"""

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model


class CoreStore(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Store`.
    """
    __tablename__ = 'corepos_store'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['store.uuid'],
                                name='corepos_store_fk_store'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    store = orm.relationship(
        model.Store,
        doc="""
        Reference to the actual store record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this store.
            """))

    corepos_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``Stores.storeID`` value for this store, within CORE-POS.
    """)

    def __str__(self):
        return str(self.store)

CoreStore.make_proxy(model.Store, '_corepos', 'corepos_id')


class CoreTax(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Tax`.
    """
    __tablename__ = 'corepos_tax'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['tax.uuid'],
                                name='corepos_tax_fk_tax'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    tax = orm.relationship(
        model.Tax,
        doc="""
        Reference to the actual tax record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this tax.
            """))

    corepos_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``taxrates.id`` value for this tax, within CORE-POS.
    """)

    def __str__(self):
        return str(self.tax)

CoreTax.make_proxy(model.Tax, '_corepos', 'corepos_id')


class CoreTender(model.Base):
    """
    CORE-specific extensions to :class:`rattail:rattail.db.model.Tender`.
    """
    __tablename__ = 'corepos_tender'
    __table_args__ = (
        sa.ForeignKeyConstraint(['uuid'], ['tender.uuid'],
                                name='corepos_tender_fk_tender'),
    )
    __versioned__ = {}

    uuid = model.uuid_column(default=None)
    tender = orm.relationship(
        model.Tender,
        doc="""
        Reference to the actual tender record, which this one extends.
        """,
        backref=orm.backref(
            '_corepos',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Reference to the CORE-POS extension record for this tender.
            """))

    corepos_id = sa.Column(sa.Integer(), nullable=False, doc="""
    ``tenders.TenderID`` value for this tender, within CORE-POS.
    """)

    def __str__(self):
        return str(self.tender)

CoreTender.make_proxy(model.Tender, '_corepos', 'corepos_id')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail-COREPOS Config Extension
"""

import warnings

from wuttjamaican.db.conf import get_engines

from rattail.config import ConfigExtension


class RattailCOREPOSExtension(ConfigExtension):
    """
    Config extension for Rattail-COREPOS
    """
    key = 'rattail-corepos'

    def configure(self, config):

        # show deprecation warnings by default, when they occur in corepos
        # (nb. rattail_corepos warnings are already shown, per rattail)
        warnings.filterwarnings('default', category=DeprecationWarning,
                                module=r'^corepos')

        # office_op
        from corepos.db.office_op import Session
        engines = get_engines(config, 'corepos.db.office_op')
        config.core_office_op_engines = engines
        config.core_office_op_engine = engines.get('default')
        Session.configure(bind=config.core_office_op_engine)
        # TODO: deprecate / remove these next 2 lines
        config.corepos_engines = engines
        config.corepos_engine = engines.get('default')

        # office_trans
        from corepos.db.office_trans import Session
        engines = get_engines(config, 'corepos.db.office_trans')
        config.core_office_trans_engines = engines
        config.core_office_trans_engine = engines.get('default')
        Session.configure(bind=config.core_office_trans_engine)
        # TODO: deprecate / remove these next 2 lines
        config.coretrans_engines = engines
        config.coretrans_engine = engines.get('default')

        # office_arch
        from corepos.db.office_arch import Session
        engines = get_engines(config, 'corepos.db.office_arch')
        if not engines:
            engines = get_engines(config, 'corepos.db.office_trans_archive')
            if engines:
                warnings.warn("config section [corepos.db.office_trans_archive] is deprecated; "
                              "please use section [corepos.db.office_arch] instead",
                              DeprecationWarning)
        config.core_office_arch_engines = engines
        config.core_office_arch_engine = engines.get('default')
        # TODO: deprecate / remove these 2
        config.core_office_trans_archive_engines = config.core_office_arch_engines
        config.core_office_trans_archive_engine = config.core_office_arch_engine
        Session.configure(bind=config.core_office_arch_engine)

        # lane_op
        from corepos.db.lane_op import Session
        engines = get_engines(config, 'corepos.db.lane_op')
        config.core_lane_op_engines = engines
        config.core_lane_op_engine = engines.get('default')
        Session.configure(bind=config.core_lane_op_engine)

        ##############################
        # import handlers
        ##############################

        # rattail corepos-import-square
        config.setdefault('rattail.importing', 'to_corepos_db_office_trans.from_square_csv.import.default_handler',
                          'rattail_corepos.corepos.importing.db.square:FromSquareToCoreTrans')
        config.setdefault('rattail.importing', 'to_corepos_db_office_trans.from_square_csv.import.default_cmd',
                          'rattail corepos-import-square')
        # TODO: there was not a legacy setting in place for this one
        # config.setdefault('rattail.importing', 'to_corepos_db_office_trans.from_square_csv.import.legacy_handler_setting',
        #                   'corepos.importing, square.handler')

        # rattail export-corepos
        config.setdefault('rattail.importing', 'to_corepos_api.from_rattail.export.default_handler',
                          'rattail_corepos.corepos.importing.rattail:FromRattailToCore')
        config.setdefault('rattail.importing', 'to_corepos_api.from_rattail.export.default_cmd',
                          'rattail export-corepos')
        config.setdefault('rattail.importing', 'to_corepos_api.from_rattail.export.legacy_handler_setting',
                          'rattail.exporting, corepos.handler')

        # rattail import-corepos-api
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_api.import.default_handler',
                          'rattail_corepos.importing.corepos.api:FromCOREPOSToRattail')
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_api.import.default_cmd',
                          'rattail import-corepos-api')
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_api.import.legacy_handler_setting',
                          'rattail.importing, corepos_api.handler')

        # rattail import-corepos-db
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_db_office_op.import.default_handler',
                          'rattail_corepos.importing.corepos.db:FromCOREPOSToRattail')
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_db_office_op.import.default_cmd',
                          'rattail import-corepos-db')
        config.setdefault('rattail.importing', 'to_rattail.from_corepos_db_office_op.import.legacy_handler_setting',
                          'rattail.importing, corepos.handler')

        # trainwreck import-corepos
        config.setdefault('rattail.importing', 'to_trainwreck.from_corepos_db_office_trans.import.default_handler',
                          'rattail_corepos.trainwreck.importing.corepos:FromCoreToTrainwreck')
        config.setdefault('rattail.importing', 'to_trainwreck.from_corepos_db_office_trans.import.default_cmd',
                          'trainwreck import-corepos')
        # TODO: there was not a legacy setting in place for this one
        # config.setdefault('rattail.importing', 'to_trainwreck.from_corepos_db_office_trans.import.legacy_handler_setting',
        #                   'trainwreck.importing, corepos.handler')

        # core-office export-lane-op
        config.setdefault('rattail.importing', 'to_corepos_db_lane_op.from_corepos_db_office_op.export.default_handler',
                          'rattail_corepos.corepos.lane.importing.op.office:FromCoreOfficeToCoreLane')
        config.setdefault('rattail.importing', 'to_corepos_db_lane_op.from_corepos_db_office_op.export.default_cmd',
                          'core-office export-lane-op')
        config.setdefault('rattail.importing', 'to_corepos_db_lane_op.from_corepos_db_office_op.export.legacy_setting',
                          'corepos.lane.importing, office.handler')

        # core-office import-self
        config.setdefault('rattail.importing', 'to_self.from_corepos_db_office_op.import.default_handler',
                          'rattail_corepos.corepos.office.importing.db.local:FromCoreOfficeToSelf')
        config.setdefault('rattail.importing', 'to_self.from_corepos_db_office_op.import.default_cmd',
                          'core-office import-self')

        # crepes export-core
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.export.default_handler',
                          'rattail_corepos.corepos.importing.db.corepos:FromCoreToCoreExport')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.export.default_cmd',
                          'crepes export-core')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.export.legacy_setting',
                          'rattail_corepos.exporting, corepos.handler')

        # crepes export-csv
        config.setdefault('rattail.importing', 'to_csv.from_corepos_db_office_op.export.default_handler',
                          'rattail_corepos.corepos.importing.db.exporters.csv:FromCoreToCSV')
        config.setdefault('rattail.importing', 'to_csv.from_corepos_db_office_op.export.default_cmd',
                          'crepes export-csv')
        config.setdefault('rattail.importing', 'to_csv.from_corepos_db_office_op.export.legacy_setting',
                          'rattail_corepos.exporting, csv.handler')

        # crepes import-core
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.import.default_handler',
                          'rattail_corepos.corepos.importing.db.corepos:FromCoreToCoreImport')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.import.default_cmd',
                          'crepes import-core')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_corepos_db_office_op.import.legacy_setting',
                          'rattail_corepos.importing, corepos.handler')

        # crepes import-csv
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_csv.import.default_handler',
                          'rattail_corepos.corepos.importing.db.csv:FromCSVToCore')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_csv.import.default_cmd',
                          'crepes import-csv')
        config.setdefault('rattail.importing', 'to_corepos_db_office_op.from_csv.import.legacy_setting',
                          'rattail_corepos.importing, csv.handler')

        ##############################
        # batch handlers
        ##############################

        # corepos_member
        config.setdefault('rattail.batch', 'corepos_member.handler.default',
                          'rattail_corepos.batch.coremember:CoreMemberBatchHandler')

        # corepos_equity_import
        config.setdefault('rattail.batch', 'corepos_equity_import.handler.default',
                          'rattail_corepos.batch.equityimport:CoreEquityImportBatchHandler')


def core_office_url(config, require=False, **kwargs):
    """ DEPRECATED """
    warnings.warn("core_office_url() function is deprecated; please use "
                  "corepos_handler.get_office_url() instead",
                  DeprecationWarning, stacklevel=2)

    app = config.get_app()
    corepos = app.get_corepos_handler()
    return corepos.get_office_url(require=require, **kwargs)


def core_office_customer_account_url(config, card_number, office_url=None):
    """ DEPRECATED """
    warnings.warn("core_office_customer_account_url() function is deprecated; please use "
                  "corepos_handler.get_office_customer_account_url() instead",
                  DeprecationWarning, stacklevel=2)

    app = config.get_app()
    corepos = app.get_corepos_handler()
    return corepos.get_office_member_url(card_number, office_url=office_url, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CORE Office utility logic
"""

import json
import os
import subprocess

from rattail.files import resource_path


def get_fannie_config_value(config, name, fannie_config_path=None):
    """
    Retrieve a config value from `fannie/config.php`

    :param config: Rattail config object.

    :param name: Name of the config value to be returned.  This need
       not be prefixed with ``FANNIE_`` although it can be.

    :param fannie_config_path: Optional path to fannie config file.
       If not specified, it is will be determined via ``config``.

    :returns: Config value as Python object, e.g. a string or dict.
       This is believed to work okay but since the Fannie config file
       represents data with PHP code, which is then converted to JSON
       and finally to Python, some discrepancies may be possible.
    """
    if not name.startswith('FANNIE_'):
        name = f'FANNIE_{name}'

    if not fannie_config_path:
        is4c = config.require('corepos', 'srcdir')
        fannie_config_path = os.path.join(is4c, 'fannie', 'config.php')
    script = resource_path('rattail_corepos.corepos.office:scripts/parse-fannie-config.php')

    try:
        output = subprocess.check_output(['php', script,
                                          '--path', fannie_config_path,
                                          '--setting', name])
    except subprocess.CalledProcessError as error:
        output = error.output.decode('utf_8')
        if "file does not contain setting" in output:
            raise ValueError(output)
        raise

    return json.loads(output.decode('utf_8'))


def set_fannie_config_value(config, name, value, fannie_config_path=None):
    """
    Update a config value in `fannie/config.php`

    :param config: Rattail config object.

    :param name: Name of the config value to set.  This need not be
       prefixed with ``FANNIE_`` although it can be.

    :param value: Config value to set This must be provided as a
       JSON-compatible string.

    :param fannie_config_path: Optional path to fannie config file.
       If not specified, it is will be determined via ``config``.
    """
    if not name.startswith('FANNIE_'):
        name = f'FANNIE_{name}'

    if not fannie_config_path:
        is4c = config.require('corepos', 'srcdir')
        fannie_config_path = os.path.join(is4c, 'fannie', 'config.php')
    script = resource_path('rattail_corepos.corepos.office:scripts/update-fannie-config.php')

    output = subprocess.check_output(['php', script,
                                      '--path', fannie_config_path,
                                      '--name', name,
                                      '--value', json.dumps(value)])


def get_blueline_template(config):
    return get_fannie_config_value(config, 'BLUELINE_TEMPLATE')


def make_blueline(config, customer, template=None):
    if not template:
        template = get_blueline_template(config)

    blueline = template
    blueline = blueline.replace('{{ACCOUNTNO}}', str(customer.card_number))
    blueline = blueline.replace('{{ACCOUNTTYPE}}', customer.member_type.description if customer.member_type else '??')
    blueline = blueline.replace('{{FIRSTNAME}}', customer.first_name or '')
    blueline = blueline.replace('{{FIRSTINITIAL}}', customer.first_name[0] if customer.first_name else '')
    blueline = blueline.replace('{{LASTNAME}}', customer.last_name or '')
    blueline = blueline.replace('{{LASTINITIAL}}', customer.last_name[0] if customer.last_name else '')
    return blueline



# -*- coding: utf-8; -*-
"""add corepos_member batch

Revision ID: 50961b4b854a
Revises: 7fea5aebddfb
Create Date: 2021-11-04 18:36:23.494783

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '50961b4b854a'
down_revision = '7fea5aebddfb'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_corepos_member
    op.create_table('batch_corepos_member',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('complete', sa.Boolean(), nullable=False),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True),
                    sa.Column('extra_data', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('input_file', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'], name='batch_corepos_member_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='batch_corepos_member_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'], name='batch_corepos_member_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_corepos_member_row
    op.create_table('batch_corepos_member_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('modified', sa.DateTime(), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('card_number', sa.Integer(), nullable=True),
                    sa.Column('first_name', sa.String(length=30), nullable=True),
                    sa.Column('first_name_old', sa.String(length=30), nullable=True),
                    sa.Column('last_name', sa.String(length=30), nullable=True),
                    sa.Column('last_name_old', sa.String(length=30), nullable=True),
                    sa.Column('street', sa.String(length=255), nullable=True),
                    sa.Column('street_old', sa.String(length=255), nullable=True),
                    sa.Column('city', sa.String(length=20), nullable=True),
                    sa.Column('city_old', sa.String(length=20), nullable=True),
                    sa.Column('state', sa.String(length=2), nullable=True),
                    sa.Column('state_old', sa.String(length=2), nullable=True),
                    sa.Column('zipcode', sa.String(length=10), nullable=True),
                    sa.Column('zipcode_old', sa.String(length=10), nullable=True),
                    sa.Column('phone', sa.String(length=30), nullable=True),
                    sa.Column('phone_old', sa.String(length=30), nullable=True),
                    sa.Column('email1', sa.String(length=50), nullable=True),
                    sa.Column('email1_old', sa.String(length=50), nullable=True),
                    sa.Column('email2', sa.String(length=50), nullable=True),
                    sa.Column('email2_old', sa.String(length=50), nullable=True),
                    sa.Column('member_type_id', sa.SmallInteger(), nullable=True),
                    sa.Column('member_type_id_old', sa.SmallInteger(), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], ['batch_corepos_member.uuid'], name='batch_corepos_member_row_fk_batch_uuid'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # batch_corepos_member*
    op.drop_table('batch_corepos_member_row')
    op.drop_table('batch_corepos_member')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Products Handler
"""

from corepos.db.office_op import Session as CoreSession

from rattail import products as base


class CoreProductsHandlerMixin(object):
    """
    Products handler mixin for CORE-POS integration
    """

    def find_wild_uoms_in_corepos(self, session, **kwargs):
        core_session = CoreSession()

        wild_uoms = core_session.execute("""
        SELECT DISTINCT UPPER(TRIM(unitofmeasure))
        FROM products
        WHERE unitofmeasure IS NOT NULL AND TRIM(unitofmeasure) != ''
        ORDER BY 1
        """).fetchall()

        core_session.close()
        return [row[0] for row in wild_uoms]


class CoreProductsHandler(base.ProductsHandler, CoreProductsHandlerMixin):
    """
    Custom products handler for use with CORE-POS. 
    """

    def find_wild_uoms(self, session, **kwargs):
        return self.find_wild_uoms_in_corepos(session, **kwargs)
